# -*- coding: utf-8 -*-
r"""
Pyvista implementation for visualizing energy surface on 2d configuration spaces.
"""
import numpy as np
import pyvista as pv
from pathlib import Path
from spinterface.visualizations.energylandscapes.ivisualizer import IVisualizer
from spinterface.visualizations.energylandscapes.i2dfunction import I2DFunction
from typing import Union, List, Tuple


class CVisualPyVista(IVisualizer):
    r"""
    Pyvista implementation for visualizing function on 2d space
    """

    def __init__(self, x: np.ndarray, y: np.ndarray, function: I2DFunction,
                 cam: Union[List[Tuple[float, float, float]], None] = None) -> None:
        r"""
        Initializes the visualization

        Args:
            x(np.array): x array - can be created with linspace (together with y a meshgrid will be created.
            y(np.array): y array - can be created with linspace (together with x a meshgrid will be created.
            function (I2DFunction): function which meets the requirements for this visualization technique
            camera: camera position, default None
        """
        super().__init__(x=x, y=y, function=function)
        self.cam = cam
        self.meshes_options, self.meshes = [], []
        self._make_plotter()

    def add_mesh(self, mesh, **kwargs):
        self.plotter.add_mesh(mesh, **kwargs)
        self.meshes_options.append(dict(**kwargs))
        self.meshes.append(mesh)

    def _make_plotter(self, offscreen: bool = False):
        r"""
        Creates the plotter. The plotter will be recreated when saving the image
        """
        self._configureplotter(offscreen)
        self.plotter.camera_position = self.cam
        self._cloud = self._make_pointcloud()
        # delaunay_
        self._surf = self._triangulation()
        # use height information for cmap and contour plots
        self._surf.point_arrays['scalars'] = self._surf.points[:, 2]
        self._cont = self._surf.contour()
        self.plotter.add_mesh(self._surf, scalars='scalars', show_scalar_bar=False)
        self.plotter.add_mesh(self._cont, color='white', show_scalar_bar=False, line_width=3.5)
        for (idx, mesh) in enumerate(self.meshes):
            self.plotter.add_mesh(mesh, **self.meshes_options[idx])

    def _triangulation(self, technique: str = 'delaunay_2d') -> pv.PolyData:
        r"""
        Creates pyvista surface from point cloud

        Args:
            technique(str): technique which is used for triangulation

        Returns:
            the triangulated surface in pyvista fashion
        """
        if technique == 'delaunay_2d':
            return self._cloud.delaunay_2d()
        else:
            raise NotImplementedError('Triangulation technique not implemented yet.')

    def _make_pointcloud(self) -> pv.PolyData:
        r"""
        Takes the points from the grid defined on the 2d conf. space together with the function values and creates
        a 3D point cloud with the requirements of pyvista.

        Returns:
            pyvista polydata object for further use
        """
        self.r = np.dstack([self.xgrid, self.ygrid, self.zgrid])
        self.rr = np.reshape(self.r, (self.r.shape[0] * self.r.shape[1], self.r.shape[2]))
        return pv.PolyData(self.rr)

    def _configureplotter(self, offscreen: bool) -> None:
        r"""
        Configures the plotter object
        """
        theme = pv.themes.ParaViewTheme()
        theme.transparent_background = True
        self.plotter = pv.Plotter(off_screen=offscreen, lighting='three lights', theme=theme)

        def cam() -> None:
            print('Camera postion: ', self.plotter.camera_position)

        self.plotter.add_key_event('c', cam)

    def __call__(self, outpath: Path = Path.cwd() / 'landscape.png'):
        r"""
        Saves the image to a file

        Args:
            outpath(Path): output path for the png image created.
        """
        self._make_plotter(offscreen=True)
        self.plotter.window_size = [4000, 4000]
        self.plotter.screenshot(str(outpath.stem))

    def show(self) -> None:
        r"""
        Shows the plotter
        """
        print('Look what you have done.......')
        print('to get current cam-position press key c')
        self.plotter.show()

    def show_grid_topview(self) -> None:
        r"""
        Shows the topview with grid to get an idea concerning the position of points
        """
        self.plotter.show_grid(color='k')
        self.plotter.view_xy()
        self.plotter.show()

    def placepointonsurface(self, xypoint: np.ndarray, color: str = 'k', point_size: float = 20) -> np.ndarray:
        r"""
        Projects a 2d point on the energy surface and add the mesh to the plotter

        Args:
            xypoint: 2d numpy array which will be projected

        Returns:
            xyz coordinates of point
        """
        if len(xypoint) != 2:
            raise ValueError("xypoint has to have length 2 (xy-coords.)")
        l_point = self.projectosurface(xypoint=xypoint)
        l_pointpolydata = pv.PolyData(l_point)
        self.add_mesh(l_pointpolydata, color=color, point_size=point_size, render_points_as_spheres=True)
        return l_point

    def projectosurface(self, xypoint: np.ndarray) -> np.ndarray:
        r"""
        Takes a 2D point, projects it on the surface (f(x,y)) and returns a 3d point
        Args:
            xypoint: array

        Returns:
            [x,y,f(x,y)]
        """
        z = self.func(x=xypoint[0], y=xypoint[1])
        return np.array([xypoint[0], xypoint[1], z])

    def placepointsequenceonsurface(self, seq_xypoint: np.ndarray, color_points: str = 'k', color_line: str = 'gray',
                                    point_size: float = 20, tuberadius: float = 0.025) -> np.ndarray:
        r"""
        Places a sequence of points on the 2d surface and connect them with a tube.

        Args:
            seq_xypoint: a sequence of 2d points: [[x1,y1],[x2,y2],...]
            color_points: the color of the points
            color_line: the color of the connecting tube
            point_size: the size of the points

        Returns:
            the sequence of points in 3D [[x1,y1,f(x1,y1)],...]
        """

        def build_path_section(start: np.ndarray, end: np.ndarray, num: int = 10) -> List[np.ndarray]:
            r"""
            gets the xy position of a start and endpoint and distributes num equally space points between them.

            Args:
                start: 3d point
                end: 3d point
                num: number of intermediate points

            Returns:
                all points
            """
            vec = end - start
            l = np.linalg.norm(vec) / float(num)
            points = []
            for i in range(num + 1):
                p2d = start.copy() + l * i * (vec / np.linalg.norm(vec))
                points.append(p2d)
            return points

        if np.shape(seq_xypoint)[1] != 2:
            raise ValueError("sequence has to have xypoints with length 2 (xy-coords.)")
        z = self.func(x=seq_xypoint[:, 0], y=seq_xypoint[:, 1])
        l_points = np.column_stack((seq_xypoint[:, 0], seq_xypoint[:, 1], z))
        l_pointpolydata = pv.PolyData(l_points)
        path = []
        for (idx, point) in enumerate(l_points[:-1]):
            path.extend(build_path_section(start=point, end=l_points[idx + 1]))
        spline = pv.Spline(path, 1000)
        spline["scalars"] = np.arange(spline.n_points)
        self.add_mesh(spline.tube(radius=tuberadius), smooth_shading=True, color=color_line)
        self.add_mesh(l_pointpolydata, color=color_points, point_size=point_size, render_points_as_spheres=True)
        return l_points

    @staticmethod
    def paraboloid(x: np.ndarray, y: np.ndarray, x0: float, y0: float, eps_x: float, eps_y: float,
                   yoffset: float) -> np.ndarray:
        r"""

        Args:
            x: x array for the 2d points
            y: y array for the 2d points
            x0,y0: zero point
            eps_x, eps_y: prefactor for the paraboloid directions x and y, if one of the eps-values is negative the
            result is a hyperbolic paraboloid.

        Returns:
            f(x) = eps_x * x^2 + eps_y * y^2
        """
        # the following does not work!
        # angle between first eigenvector and the x axis
        # phi = np.arccos(evec[0]/np.linalg.norm(evec))
        # return eps_x * (np.cos(phi) *(x-x0)-np.sin(phi * (y-y0))) ** 2 + eps_y * (np.sin(phi) * (x-x0) + np.cos(phi) * (y-y0)) ** 2 + yoffset
        return eps_x * (x - x0) ** 2 + eps_y * (y - y0) ** 2 + yoffset

    def create_eigenvalue_parabola(self, point: np.ndarray, evals: np.ndarray, evecs: np.ndarray, nr_points: int = 10,
                                   point_distance_xy: float = 1.0, color_points: str = 'k', color_line: str = 'gray',
                                   point_size: float = 20, tuberadius: float = 0.025, show_surface: bool = True,
                                   x_width: float = 2,
                                   y_width: float = 2, y_offset: float = 0.0) -> None:
        r"""
        Creates the two parabolas representing the diagonalization of the hessian for a given state. This means
        visualizing a parabola (with curvature of the eigenvalue eps) in the direction of each eigenvector
        f(p) = eps * dp + p0

        with dp = |p-p0| and p0 the initial point

        Args:
            show_surface: boolean whether the paraboloic surface will be visualized as a surface
            color_line: color of the line parabola along the certain eigenvector
            color_points: color of the points along the certain eigenmode
            tuberadius: radius of the line
            point_size: size of the points
            nr_points: number of points for which the parabola is evaluated
            point_distance_xy: the xy-distance between the points
            evecs: the eigenvectors of the "point"-state
            point: the points which eigenvectors will be visualized
            evals: the eigenvalues of the "point"-state
        """
        # create points
        for (idx, evec) in enumerate(evecs):
            eval = evals[idx]
            evec = evec / np.linalg.norm(evec)
            points_positive, energy_positive = [], []
            for n in range(nr_points):
                points_positive.append(point + point_distance_xy * n * evec)
                # calc energy value (leaving the energy surface)
                energy_positive.append(
                    self.func(x=point[0], y=point[1]) + eval * (point_distance_xy * n) ** 2 + y_offset)
            points_positive.reverse()
            energy_positive.reverse()
            # mirror the parabola
            points_negative, energy_negative = [], []
            for n in range(nr_points):
                points_negative.append(point - point_distance_xy * n * evec)
                # calc energy value (leaving the energy surface)
                energy_negative.append(
                    self.func(x=point[0], y=point[1]) + eval * (point_distance_xy * n) ** 2 + y_offset)
            points = np.asarray(points_positive + points_negative)
            energy = np.asarray(energy_positive + energy_negative)
            l_points = np.column_stack((points[:, 0], points[:, 1], energy))
            l_pointpolydata = pv.PolyData(l_points)
            spline = pv.Spline(l_points, 1000)
            spline["scalars"] = np.arange(spline.n_points)
            self.add_mesh(spline.tube(radius=tuberadius), smooth_shading=True, color=color_line)
            self.add_mesh(l_pointpolydata, color=color_points, point_size=point_size, render_points_as_spheres=True)
            if show_surface:
                xx = np.linspace(point[0] - x_width / 2, point[0] + x_width / 2, 20)
                yy = np.linspace(point[1] - y_width / 2, point[1] + y_width / 2, 20)
                xgrid, ygrid = np.meshgrid(xx, yy)
                # evaluate the
                zgrid = self.paraboloid(xgrid, ygrid, x0=point[0], y0=point[1], eps_x=evals[0], eps_y=evals[1],
                                        yoffset=float(self.func(point[0], point[1])) + y_offset)
                points = np.dstack([xgrid, ygrid, zgrid])
                print(points)
                r_points = pv.PolyData(np.reshape(points, (points.shape[0] * points.shape[1], points.shape[2])))
                surf = r_points.delaunay_2d().triangulate()
                self.add_mesh(surf, color=color_line, show_scalar_bar=False)

    @staticmethod
    def create_sequence_between_points(initial: np.ndarray, final: np.ndarray, nr: int) -> np.ndarray:
        r"""
        Creates a sequence of (nr) points on a direct 2d connection between the initial and final point.

        Args:
            initial: start point (2D)
            final: final point
            nr: number of points

        Returns:
            the creates sequence
        """
        r = (final - initial) / nr
        seq = []
        for n in range(nr + 1):
            seq.append(initial + n * r)
        return np.asarray(seq)



# -*- coding: utf-8 -*-
r"""
Mode following technique for two dimensional functions.
"""
import numpy as np
from pathlib import Path
from typing import Callable, Tuple, List, Union

import pandas as pd


class CModeFollowing2D:
    r"""

    """

    def __init__(self, f: Callable, initial: np.ndarray, mf_steps: int = 100, mf_steplength: float = 0.1,
                 h: float = 0.00001) -> None:
        r"""
        Initialize mode follower.

        Args:
            f(Callable): The energy function describing the surface

            initial(np.ndarray): 2D point (will be automatically projected on the surface). For this configuration the
            hessian will be diagonalized).

            h(float): finite element for hessian approximations

            mf_steps(int): number of iterations following one mode

            mf_steplength(float): scaling factor for the displacement along the eigenvector
        """
        self._initial = initial
        self._f = self._function_wrapper(f=f)
        self._h = h
        self._mfsteplength = mf_steplength
        self._mfsteps = mf_steps
        # identify the eigenvectors and eigenvalues for the initial state
        self._eigenvalues_initial, self._eigenvectors_initial = self.diagonalize_hessian(self._initial)
        print(f'The given initial state has the eigenvalues: {self._eigenvalues_initial} \n The given intial state'
              f'has the eigenvectors: {self._eigenvectors_initial}.')

    @property
    def eigenvalues_initial(self) -> np.ndarray:
        r"""
        Returns:
            the eigenvalues of the initial point
        """
        return self._eigenvalues_initial

    @property
    def eigenvectors_initial(self) -> np.ndarray:
        r"""
        Returns:
            the eigenvectors of the intial point
        """
        return self._eigenvectors_initial

    @staticmethod
    def _function_wrapper(f: Callable) -> Callable:
        r"""
        This wrapper changes the input structure of the callable to be in line with the visualization interface. The
        visualization interface uses a function which takes f(x,y) while the methods in this class need a callable with
        the parameter f(r(x,y)). This wrapper provides this functionality.
        Args:
            f: takes a callable with the functionality f(x,y)

        Returns: a function with f(r).
        """

        def wrapper(r: np.ndarray) -> np.ndarray:
            if len(np.shape(r)) == 1:
                x = r[0]
                y = r[1]
            else:
                x = r[:, 0]
                y = r[:, 1]
            return f(x, y)

        return wrapper

    def __call__(self, mode_index: int = 0, i_write_out: bool = False, gralog: int = 1,
                 write_out: Union[Path, None] = None) -> Tuple[List[np.ndarray],List[float],List[np.ndarray]]:
        r"""

        Args:
            mode_index(int): index of the mode to follow
            i_write_output(bool): boolean whether an output is written to a file
            gralog(int): frequency for output writing
            write_out(Path): Path for potential output file

        Returns:
            a list of the points (2D ndarrays), the eigenvalues and the eigenvectors calculated for the mode following
            for the selected mode
        """
        if write_out is None:
            write_out = Path.cwd() / f'mf_{mode_index}.dat'
        if mode_index not in [0, 1]:
            raise ValueError("This is a 2D problem. There are only 2 evecs. Please enter 0 or 1.")

        evec = self._eigenvectors_initial[mode_index]
        eval = self._eigenvalues_initial[mode_index]
        current_point = self._initial
        write_dict = {'iteration': [0], 'current_point': [current_point], 'eigenvalue': [eval],
                          'eigenvector': [evec]}

        for step in range(self._mfsteps):
            displaced_point = self._displace_structure(current_point, evec, self._mfsteplength)
            new_evals, new_evecs = self.diagonalize_hessian(displaced_point)
            new_index, evec = self._modetracker(evec, new_evecs)
            if np.mod(step, gralog) == 0:
                write_dict['iteration'].append(step)
                write_dict['current_point'].append(displaced_point)
                write_dict['eigenvalue'].append(new_evals[new_index])
                write_dict['eigenvector'].append(evec)
                df = pd.DataFrame(write_dict)
                df.to_json(write_out)
            current_point = displaced_point
        if i_write_out:
            df = pd.DataFrame(write_dict)
            df.to_json(write_out)
        return write_dict['current_point'], write_dict['eigenvalue'], write_dict['eigenvector']


    def _modetracker(self, current_eigenvector: np.ndarray, new_eigenvectors: np.ndarray) -> Tuple[int, np.ndarray]:
        r"""
        Chooses the eigenvector from the new eigenvector set which gives the highest normalized inner product with the
        previous eigenvector.

        Args:
            current_eigenvector(np.ndarray): the current eigenvector

        Returns:
            the best choice for the eigenvector to stay on the same mode and the index of the eigenvalue to which the
            new eigenvector belongs to
        @todo: something seems not to work here! (For the sp it always detects the same mode)
        """
        best_quality = 0  # the closer to 1 the better
        chosen_evec = None
        index_chosen_evec = None
        for (idx, evec) in enumerate(new_eigenvectors):
            quality = np.dot(evec, current_eigenvector) / (np.linalg.norm(evec) * np.linalg.norm(current_eigenvector))

            if abs(quality) >= best_quality:
                if quality<=0:
                    chosen_evec = -1 *evec
                else:
                    chosen_evec = evec
                index_chosen_evec = idx
        return index_chosen_evec, chosen_evec

    @staticmethod
    def _displace_structure(point: np.ndarray, displacement: np.ndarray, steplength: float = 0.1) -> np.ndarray:
        r"""
        Displaces a 2D point along a 2D vectors according to a steplength

        Args:
            point: point for displacement (usually the current step of the modefollowing
            displacement: displacementvector (2D) (usually the current eigenvector
            steplength: scale the step

        Returns:
            the displaced point
        """
        return point + displacement * steplength

    def diagonalize_hessian(self, point: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Diagonalize the hessian of the energy landscape for a given point.

        Args:
            point: 2D point for which the hessian will be diagonalized

        Returns:
            the eigenvalues and eigenvectors
        """
        hess = self._hessian(point, self._h)
        # calculate eigenvalues and normalized eigenvectors:
        return np.linalg.eigh(hess)

    def _hessian(self, point: np.ndarray, h: float) -> np.ndarray:
        r"""
        Calculate the hessian for a given point with finite elements

        Args:
            point(np.ndarray): Point for which the hessian is evaluated.
            h(float): finite element

        Returns:
            numpy array of hessian: [[dfdxdx, dfdxdy],[dfdydx, dfdydy]]
        """
        hess = np.zeros(shape=(2, 2))
        # diagonal terms
        hess[0, 0] = (self._f(point + 2 * h * np.array([1, 0])) - 2 * self._f(point + h * np.array([1, 0])) + self._f(
            point)) / h ** 2
        hess[1, 1] = (self._f(point + 2 * h * np.array([0, 1])) - 2 * self._f(point + h * np.array([0, 1])) + self._f(
            point)) / h ** 2
        hess[0, 1] = (self._f(point + h * np.array([1, 1])) - self._f(point + h * np.array([1, 0])) - self._f(
            point + h * np.array([0, 1])) + self._f(point)) / h ** 2
        hess[1, 0] = hess[0, 1]
        return hess

    def _gradient(self, point: np.ndarray, h: float) -> np.ndarray:
        r"""
        Calculate the energy gradient

        Args:
            point(np.ndarray): Point for which the gradient is evaluated.
            h(float): finite element

        Return:
             1d numpy array with length 2: [dfdx,dfdy]
        """
        dfdx = (self._f(point + h * np.array([1, 0])) - self._f(point)) / h
        dfdy = (self._f(point + h * np.array([0, 1])) - self._f(point)) / h
        return np.array([dfdx, dfdy])



from pathlib import Path
from spinterface.visualizations.lattices.cvisualpyvistabilayer import CVisualPyvistaBilayer
from spinterface.inputs.lattice.CLattice import CLattice
from spinterface.visualizations.lattices.cvisualpyvista import CVisualPyVista

latt = CLattice(source='STM', path = Path.cwd() / 'sp_J_AA0.0_im.dat')

# visualizer = CVisualPyVista(lattice=latt)
visualizer = CVisualPyvistaBilayer(lattice=latt,cmap='coolwarm',filter_around_layercenter=7, filter_half=True, dist_bot=15, dist_top=15)
#visualizer()
visualizer.show()


# -*- coding: utf-8 -*-
r"""
This module takes care of writing input files for calculations with the spindynamic program. To achieve forward
compatability you can define all the key value pairs while initializing this class
"""
import shutil
from typing import TypeVar, List
from pathlib import Path
import fileinput
import os


class CWriteInput:
    r"""
    General class for writing input files
    """

    def __init__(self, name: str, lines: List[str] = [], **kwargs) -> None:
        r"""
        Initializes writing of input file. The key value pairs of the arguments are passed towards the input file.
        To create the file you need to call this class

        Args:
            name(str): name of the created file
            lines: a list of lines which shall be added to the input file. This can be used if there are lines without
            a specific key
        """
        self.file = name
        if os.path.isfile(self.file):
            print(f'file {name} already exists. Will replace it.')
            os.remove(self.file)
        self.content = dict(**kwargs)
        self.contentlines = lines

    def __call__(self, where: Path = Path.cwd()) -> None:
        r"""
        Calls the writing

        Args:
            where(Path): In which directory shall the file be created with the name given in the initializer
        """
        self.where = where
        with open(str(where / self.file), 'a') as f:
            for (key, value) in self.content.items():
                f.write(key + ' ' + value + '\n')
            for line in self.contentlines:
                f.write(line + '\n')

    def appendline(self, value: str) -> None:
        r"""
        Appends a line to the input file
        """
        self.contentlines.append(value)

    def adjustparameter(self, key: str, value: str) -> None:
        r"""
        Adjust the parameter in the object. You have to call this object again to write.

        Args:
            key: str
            value: str
        """
        try:
            self.content[key] = value
        except KeyError:
            print(f'Key ({key}) not found in {self.file}. Will ignore command.')



r"""
Module for visualizing the tree for a whole sps-calculation
"""
from spinterface.sps_results.stage_visualizer.ivisstage import IVisStage
from pathlib import Path
from typing import Union, Dict
from treelib import Node, Tree
import ete3 as et
import pandas as pd
from decimal import Decimal
import math as mt


class CVisTree(IVisStage):
    r"""
    Visualizes all the stages of a sps-calculation in a tree.
    """

    def __init__(self, strategy: int, info_file_disp: Union[Path, str], info_file_esc: Union[Path, str],
                 info_file_conv: Union[Path, str],
                 stage_strategies: Dict[str, int] = {'DISP': 2, 'ESC': 1, 'CONV': 1, 'ESCA': 1}) -> None:
        r"""
        Initialize the visualization stage
        Args:
            strategy(int): for each stage a certain strategy was applied. The visualization also depends on these strategies.
            info_file_conv(Path,str): information about the convergence stage
            info_file_disp(Path,str): information about the dipslace stage
            info_file_esc(Path,str): infomration about the escape stage
        """
        self.strategy = strategy
        self.strategies_stages = stage_strategies
        self.abbreviations_filenames = self._setup_abbreviations()
        self.info_file_disp = info_file_disp
        self.info_file_esc = info_file_esc
        self.info_file_conv = info_file_conv
        self.parent_directory = self.info_file_conv.parent
        self.image_scale_initial = 500
        self.image_scale_disp = 500
        self.image_scale_esc = 500
        self.image_scale_conv = 500
        self.branchlength = 0.1
        self.branchlength_disp = 0.1


    def _setup_abbreviations(self) -> Dict[str, str]:
        r"""
        :return: Depending on the chosen algorithm the files are named differently. This method assigns the correct
        abbreviations depending on the algorithms for the corresponding stages.
        """
        l_abb={}
        if self.strategies_stages['ESC'] == 1:
            l_abb['ESC'] = 'mf'
        if self.strategies_stages['CONV'] == 1:
            l_abb['CONV'] = 'mmf'
        elif self.strategies_stages['CONV'] == 2:
            l_abb['CONV'] = 'mfef'
        return l_abb

    def set_image_scales(self, initial: float = 500, disp: float = 500, esc: float = 500, conv: float = 500):
        r"""
        Sets the sizes of the images for the different stages
        """
        self.image_scale_conv = conv
        self.image_scale_disp = disp
        self.image_scale_initial = initial
        self.image_scale_esc = esc

    @property
    def stage(self) -> str:
        r"""
        Returns:
            the name of the stage
        """
        return 'Visualize Tree'

    def _stagestrategytext(self) -> str:
        r"""
        Returns:
            the strategy of the stage
        """
        if self.strategy in [1]:
            return 'Simple Visualization with treelib'
        elif self.strategy in [2]:
            return 'Advanced Visualization with ete3'
        else:
            raise ValueError('Visualization strategy for tree not yet coded.')

    def __call__(self) -> None:
        r"""
        Calls the certain visualization.
        """
        if self.strategy in [1]:
            self._create_tree_treelib()
        if self.strategy in [2]:
            self._create_tree_ete()

    def _create_newick_string(self) -> str:
        r"""
        Creates a newick string describing the topology and the nodes of the tree. Also included are basic information
        as the branch distances.
        """
        df_disp = pd.read_csv(self.info_file_disp, sep=r'\s+')
        df_esc = pd.read_csv(self.info_file_esc, sep=r'\s+')
        df_conv = pd.read_csv(self.info_file_conv, sep=r'\s+')
        str = ''
        for index, key in enumerate(df_disp['DISPKEY']):
            str = str + '('
            current_esc_keys = df_esc[df_esc['DISPKEY'] == key]['ESCKEY'].to_numpy()
            current_esc_folders = df_esc[df_esc['DISPKEY'] == key]['folder'].to_numpy()
            esckeys_reescape = []
            for index2, esckey in enumerate(current_esc_keys):
                splitted = esckey.split('DIR')
                if len(splitted[1]) > 1:
                    esckeys_reescape.append(esckey)
            for index2, esckey in enumerate(current_esc_keys):
                if esckey in esckeys_reescape:
                    continue
                path_to_esc_folder = current_esc_folders[index2]
                l_abb = self.abbreviations_filenames['ESC']
                df_mf_calc = pd.read_csv(self.parent_directory / Path(path_to_esc_folder) / f'info_{l_abb}.dat', r'\s+')
                eig1 = abs(df_mf_calc['eig001'].to_numpy()[-1] * 1000)  # display in meV
                df_conv_filtered = df_conv[df_conv['DISPKEY'] == key]
                try:
                    sp_or_convex = df_conv_filtered[df_conv_filtered['ESCKEY'] == esckey]['sp_or_convex'].to_numpy()[0]
                    folder_search = df_conv_filtered[df_conv_filtered['ESCKEY'] == esckey]['folder'].to_numpy()[0]
                    l_abb = self.abbreviations_filenames['CONV']
                    df_mmf = pd.read_csv(self.parent_directory / Path(folder_search) / f'info_{l_abb}.dat', sep=r'\s+')
                    eig1_conv = abs(df_mmf['eig001'].to_numpy()[-1] * 1000)  # in meV
                    if sp_or_convex == 'sp':
                        # str = str + f'({sp_or_convex}:{eig1_conv}){esckey}:{abs(eig1)},'
                        str = str + f'({sp_or_convex}:{self.branchlength}){esckey}:{self.branchlength_disp},'
                    else:
                        # look for re-escapes
                        current_reesc_keys = []
                        current_reesc_keys_attempts = []
                        for reesc_key in esckeys_reescape:
                            if reesc_key.startswith(esckey):
                                current_reesc_keys.append(reesc_key)
                                current_reesc_keys_attempts.append(int(reesc_key[-2:]))
                        if not current_reesc_keys:
                            # list is empty no successful re-escape attempt for this key
                            #str = str + f'({sp_or_convex}:{eig1_conv}){esckey}:{abs(eig1)},'
                            str = str + f'({sp_or_convex}:{self.branchlength}){esckey}:{self.branchlength_disp},'
                        else:
                            # check ordering of keys
                            current_reesc_keys = [reesc_key for _, reesc_key in
                                                  sorted(zip(current_reesc_keys_attempts, current_reesc_keys))]
                            # go through re-escape keys
                            reesc_str = '(' * (2 * (len(current_reesc_keys) - 1) + 1)
                            for reesc_key in reversed(current_reesc_keys):
                                folder = self.parent_directory / Path(
                                    df_esc[df_esc['ESCKEY'] == reesc_key]['folder'].to_numpy()[0])
                                df_esca = pd.read_csv(folder / 'info_esca.dat', sep=r'\s+')
                                eig_esca1 = abs(df_esca['eig001'].to_numpy()[-1] * 1000)
                                # look for corresponding search calculation
                                sp_or_convex = \
                                    df_conv_filtered[df_conv_filtered['ESCKEY'] == reesc_key][
                                        'sp_or_convex'].to_numpy()[0]
                                reesc_str = reesc_str + f'{sp_or_convex}:{self.branchlength}){reesc_key}:{self.branchlength})'
                            reesc_str = reesc_str[:-1]  # remove last bracket
                            # str = str + f'(({reesc_str}){sp_or_convex}:{eig1_conv}){esckey}:{abs(eig1)},'
                            str = str + f'(({reesc_str}){sp_or_convex}:{self.branchlength}){esckey}:{self.branchlength_disp},'
                except IndexError:
                    # str = str + f'(NC:{0.0}){esckey}:{abs(eig1)},'
                    str = str + f'(NC:{0.0}){esckey}:{self.branchlength_disp},'
            # remove last comma
            str = str[:-1]
            if not str[-1] == ',':
                str = str + ')'
            str = str + f'{key}:{0.05},'
        return str[:-1]

    def _design_node_initial_state(self, tree: et.Tree) -> None:
        r"""
        Add faces to root node (initial state)
        """
        # maximum number of eigenvalues created by spinaker
        max_eval = 9
        root_node = tree.search_nodes(name='initial')[0]
        if (self.parent_directory / 'SpinSTMi.png').is_file():
            img_face = et.ImgFace("SpinSTMi.png", width=self.image_scale_initial, height=self.image_scale_initial)
            root_node.add_face(img_face, column=0, position='branch-right')
        rootnodestyle = et.NodeStyle()
        rootnodestyle['bgcolor'] = "Moccasin"
        root_node.set_style(rootnodestyle)
        if (self.parent_directory / 'info_sps_eigenvalues.dat').is_file():
            df_sps_eval = pd.read_csv(self.parent_directory / 'info_sps_eigenvalues.dat', sep=r'\s+')
            if 'initialstate' in df_sps_eval['Stage'].values:
                for i in range(1, max_eval + 1):
                    eval = float(
                        str(df_sps_eval[df_sps_eval['Stage'] == 'initialstate'][f'eig{i}'].to_numpy()[0]).replace('d',
                                                                                                                  'e'))
                    str_scientific = '{:.2e}'.format(eval)
                    tf = et.TextFace(f'eig{i}=' + str_scientific, fsize=20)
                    root_node.add_face(tf, column=0, position='branch-bottom')

    def _design_node_displaced_state(self, node: et.Tree, df_disp: pd.DataFrame, index: int, key: str) -> None:
        r"""
        Add faces to displace node
        """
        max_eval = 9
        disp_spinfile = self.parent_directory / df_disp['spin_file'].to_numpy()[index]
        disp_spinfile = disp_spinfile.as_posix().replace('.dat', '.png')
        if Path(disp_spinfile).is_file():
            img_face = et.ImgFace(disp_spinfile, width=self.image_scale_disp, height=self.image_scale_disp)
            node.add_face(img_face, column=0)
        dispnodestyle = et.NodeStyle()
        dispnodestyle['bgcolor'] = "LightSteelBlue"
        node.set_style(dispnodestyle)
        if (self.parent_directory / 'info_sps_eigenvalues.dat').is_file():
            df_sps_eval = pd.read_csv(self.parent_directory / 'info_sps_eigenvalues.dat', sep=r'\s+')
            if key in df_sps_eval['Stage'].values:
                for i in range(1, max_eval + 1):
                    eval = float(
                        str(df_sps_eval[df_sps_eval['Stage'] == key][f'eig{i}'].to_numpy()[0]).replace('d', 'e'))
                    str_scientific = '{:.2e}'.format(eval)
                    tf = et.TextFace(f'eig{i}=' + str_scientific, fsize=20)
                    node.add_face(tf, column=0, position='branch-bottom')

    def _design_node_escaped_state(self, node: et.Tree, df_esc: pd.DataFrame) -> None:
        r"""
        Designs and add faces to nodes of the escape stage
        """
        current_esc_folder = df_esc[df_esc['ESCKEY'] == node.name]['folder'].to_numpy()[0]
        current_esc_spinfile = df_esc[df_esc['ESCKEY'] == node.name]['spin_file'].to_numpy()[0]
        path_to_spin_file = self.parent_directory / Path(current_esc_spinfile)
        esc_spinfile = path_to_spin_file.as_posix().replace('.dat', '.png')
        if Path(esc_spinfile).is_file():
            img_face = et.ImgFace(esc_spinfile, width=self.image_scale_esc, height=self.image_scale_esc)
            tf = et.TextFace('Escaped config.', fsize=25)
            tf.inner_border.width = 1
            tf.inner_border.type = 0
            img_face.inner_border.width = 1
            img_face.inner_border.type = 0
            node.add_face(tf, column=1, position='branch-top')
            node.add_face(img_face, column=1, position='branch-top')
        escnodestyle = et.NodeStyle()
        escnodestyle['bgcolor'] = "Khaki"
        node.set_style(escnodestyle)
        # Test if an eigenvector file for the beginning of mmf exists
        l_abb_E = self.abbreviations_filenames['ESC']
        l_abb_C = self.abbreviations_filenames['CONV']
        esc_evec0file = esc_spinfile.replace(f'spin_{l_abb_E}_end', f'evec_{l_abb_C}_0')
        if Path(esc_evec0file).is_file():
            tf = et.TextFace('Lowest eigenmode', fsize=25)
            tf.inner_border.width = 1
            tf.inner_border.type = 0
            node.add_face(tf, column=1, position='branch-bottom')
            img_face = et.ImgFace(esc_evec0file, width=self.image_scale_esc, height=self.image_scale_esc)
            img_face.inner_border.width = 1
            img_face.inner_border.type = 0
            node.add_face(img_face, column=1, position='branch-bottom')

        path_to_mf_file = self.parent_directory / Path(current_esc_folder) / f'info_{l_abb_E}.dat'
        df_mf = pd.read_csv(path_to_mf_file, sep=r'\s+')
        eig1 = df_mf['eig001'].to_numpy()[-1]
        eig2 = df_mf['eig002'].to_numpy()[-1]
        node.add_feature('eig2', eig2)
        node.add_feature('eig1', eig1)
        scientific_str = '{:.2e}'.format(float(eig1))
        tf1 = et.faces.TextFace(f"eig1:{scientific_str}", fsize=20)
        scientific_str = '{:.2e}'.format(float(eig2))
        tf2 = et.faces.TextFace(f"eig2:{scientific_str}", fsize=20)
        node.add_face(tf1, column=0, position='branch-bottom')
        node.add_face(tf2, column=0, position='branch-bottom')

    def _design_node_converged_state(self, node: et.Tree, df_conv: pd.DataFrame) -> None:
        r"""
        Designs and add faces to convergence stage
        """
        if not (df_conv['folder'].to_numpy().size == 0):
            folder_search = df_conv['folder'].to_numpy()[0]
            l_abb_C = self.abbreviations_filenames['CONV']
            df_mmf = pd.read_csv(self.parent_directory / Path(folder_search) / f'info_{l_abb_C}.dat', sep=r'\s+')
            eig1 = df_mmf['eig001'].to_numpy()[-1]
            eig2 = df_mmf['eig002'].to_numpy()[-1]
            scientific_str = '{:.2e}'.format(float(eig1))
            tf1 = et.faces.TextFace(f"eig1:{scientific_str}", fsize=20)
            scientific_str = '{:.2e}'.format(float(eig2))
            tf2 = et.faces.TextFace(f"eig2:{scientific_str}", fsize=20)
            node.add_face(tf1, column=0, position='branch-top')
            node.add_face(tf2, column=0, position='branch-top')
            energy_sp = df_mmf['energy'].to_numpy()[-1]
            steps_sp = df_mmf['steps'].to_numpy()[-1]
            scientific_str = '{:.2e}'.format(float(energy_sp))
            tf1 = et.faces.TextFace(f"energy={scientific_str} eV", fsize=20)
            tf2 = et.faces.TextFace(f"steps={steps_sp}", fsize=20)
            node.add_face(tf1, column=0, position='branch-bottom')
            node.add_face(tf2, column=0, position='branch-bottom')
            if not (df_conv['spin_file'].to_numpy().size == 0):
                spinfile_search = df_conv['spin_file'].to_numpy()[0]
                path_to_spin_file = self.parent_directory / Path(spinfile_search)
                conv_spinfile = path_to_spin_file.as_posix().replace('.dat', '.png')
                if Path(conv_spinfile).is_file():
                    img_face = et.ImgFace(conv_spinfile, width=self.image_scale_conv, height=self.image_scale_conv)
                    node.add_face(img_face, column=1, position='branch-bottom')
                convnodestyle = et.NodeStyle()
                if df_conv['sp_or_convex'].to_numpy()[0] == 'sp':
                    node.add_face(et.faces.TextFace('Unstable Mode', fsize=25), column=1, position='branch-top')
                    convnodestyle['bgcolor'] = "DarkSeaGreen"
                else:
                    node.add_face(et.faces.TextFace('Lowest Mode', fsize=25), column=1, position='branch-top')
                    convnodestyle['bgcolor'] = "LightSalmon"
                node.set_style(convnodestyle)
                conv_evecfile = conv_spinfile.replace('spin', 'evec')
                if Path(conv_evecfile).is_file():
                    img_face = et.ImgFace(conv_evecfile, width=self.image_scale_conv, height=self.image_scale_conv)
                    node.add_face(img_face, column=1, position='branch-top')

    def _design_node_reescape_esc_state(self, node: et.Tree, df_esc: pd.DataFrame) -> None:
        r"""
            Designs and add faces to nodes of the escape stage
        """
        current_esc_folder = df_esc[df_esc['ESCKEY'] == node.name]['folder'].to_numpy()[0]
        current_esc_spinfile = df_esc[df_esc['ESCKEY'] == node.name]['spin_file'].to_numpy()[0]
        path_to_spin_file = self.parent_directory / Path(current_esc_spinfile)
        esc_spinfile = path_to_spin_file.as_posix().replace('.dat', '.png')
        reescape_escnodestyle = et.NodeStyle()
        reescape_escnodestyle['bgcolor'] = "Khaki"
        node.set_style(reescape_escnodestyle)
        if Path(esc_spinfile).is_file():
            img_face = et.ImgFace(esc_spinfile, width=self.image_scale_esc, height=self.image_scale_esc)
            tf = et.TextFace('Escaped config.', fsize=25)
            tf.inner_border.width = 1
            tf.inner_border.type = 0
            img_face.inner_border.width = 1
            img_face.inner_border.type = 0
            node.add_face(tf, column=1, position='branch-top')
            node.add_face(img_face, column=1, position='branch-top')
        # Test if an eigenvector file for the beginning of mmf exists
        l_abb_C = self.abbreviations_filenames['CONV']
        path_to_evec_file = self.parent_directory / Path(current_esc_spinfile).parent / f'evec_{l_abb_C}_0.png'
        esc_evec0file = path_to_evec_file.as_posix()
        if Path(esc_evec0file).is_file():
            tf = et.TextFace('Lowest eigenmode', fsize=25)
            tf.inner_border.width = 1
            tf.inner_border.type = 0
            node.add_face(tf, column=1, position='branch-bottom')
            img_face = et.ImgFace(esc_evec0file, width=self.image_scale_esc, height=self.image_scale_esc)
            img_face.inner_border.width = 1
            img_face.inner_border.type = 0
            node.add_face(img_face, column=1, position='branch-bottom')

        path_to_esca_file = self.parent_directory / Path(current_esc_folder) / 'info_esca.dat'
        df_mf = pd.read_csv(path_to_esca_file, sep=r'\s+')
        eig1 = df_mf['eig001'].to_numpy()[-1]
        eig2 = df_mf['eig002'].to_numpy()[-1]
        node.add_feature('eig2', eig2)
        node.add_feature('eig1', eig1)
        scientific_str = '{:.2e}'.format(float(eig1))
        tf1 = et.faces.TextFace(f"eig1:{scientific_str}", fsize=20)
        scientific_str = '{:.2e}'.format(float(eig2))
        tf2 = et.faces.TextFace(f"eig2:{scientific_str}", fsize=20)
        node.add_face(tf1, column=0, position='branch-bottom')
        node.add_face(tf2, column=0, position='branch-bottom')

    def _design_node_reescape_conv_state(self, node: et.Tree, df_conv: pd.DataFrame) -> None:
        r"""
        Designs and adds faces to nodes of the re-escape stage in the convergence phases
        """
        if not (df_conv['folder'].to_numpy().size == 0):
            folder_search = df_conv['folder'].to_numpy()[0]
            l_abb_C = self.abbreviations_filenames['CONV']
            df_mmf = pd.read_csv(self.parent_directory / Path(folder_search) / f'info_{l_abb_C}.dat', sep=r'\s+')
            eig1 = df_mmf['eig001'].to_numpy()[-1]
            eig2 = df_mmf['eig002'].to_numpy()[-1]
            scientific_str = '{:.2e}'.format(float(eig1))
            tf1 = et.faces.TextFace(f"eig1:{scientific_str}", fsize=20)
            scientific_str = '{:.2e}'.format(float(eig2))
            tf2 = et.faces.TextFace(f"eig2:{scientific_str}", fsize=20)
            node.add_face(tf1, column=0, position='branch-top')
            node.add_face(tf2, column=0, position='branch-top')
            energy_sp = df_mmf['energy'].to_numpy()[-1]
            steps_sp = df_mmf['steps'].to_numpy()[-1]
            scientific_str = '{:.2e}'.format(float(energy_sp))
            tf1 = et.faces.TextFace(f"energy={scientific_str} eV", fsize=20)
            tf2 = et.faces.TextFace(f"steps={steps_sp}", fsize=20)
            node.add_face(tf1, column=0, position='branch-bottom')
            node.add_face(tf2, column=0, position='branch-bottom')
            if not (df_conv['spin_file'].to_numpy().size == 0):
                spinfile_search = df_conv['spin_file'].to_numpy()[0]
                path_to_spin_file = self.parent_directory / Path(spinfile_search)
                conv_spinfile = path_to_spin_file.as_posix().replace('.dat', '.png')
                reescape_convnodestyle = et.NodeStyle()
                if df_conv['sp_or_convex'].to_numpy()[0] == 'sp':
                    reescape_convnodestyle['bgcolor'] = "DarkSeaGreen"
                else:
                    reescape_convnodestyle['bgcolor'] = "LightSalmon"
                node.set_style(reescape_convnodestyle)
                if Path(conv_spinfile).is_file():
                    img_face = et.ImgFace(conv_spinfile, width=self.image_scale_conv, height=self.image_scale_conv)
                    node.add_face(img_face, column=1, position='branch-bottom')
                conv_evecfile = conv_spinfile.replace('spin', 'evec')
                if Path(conv_evecfile).is_file():
                    if df_conv['sp_or_convex'].to_numpy()[0] == 'sp':
                        node.add_face(et.faces.TextFace('Unstable Mode', fsize=25), column=1, position='branch-top')
                    else:
                        node.add_face(et.faces.TextFace('Lowest Mode', fsize=25), column=1, position='branch-top')
                    img_face = et.ImgFace(conv_evecfile, width=self.image_scale_conv, height=self.image_scale_conv)
                    node.add_face(img_face, column=1, position='branch-top')

    def _create_tree_ete(self):
        r"""
        Create a more advanced version of a tree using ETE
        """

        def my_layout(node):
            # if node.is_root():
            #    img_face = et.ImgFace("SpinSTMi.png", width=250, height=250)
            #    et.faces.add_face_to_node(img_face, node, column=0, position="branch-top")
            if node.is_leaf():
                # If terminal node, draws its name
                name_face = et.AttrFace("name", fsize=30)
                et.faces.add_face_to_node(name_face, node, column=0, position="branch-right")
            else:
                # If internal node, draws label with smaller font size
                name_face = et.AttrFace("name", fsize=20)
                et.faces.add_face_to_node(name_face, node, column=0, position="branch-right")
            # Adds the name face to the image at the preferred position

        df_disp = pd.read_csv(self.info_file_disp, sep=r'\s+')
        df_esc = pd.read_csv(self.info_file_esc, sep=r'\s+')
        df_conv = pd.read_csv(self.info_file_conv, sep=r'\s+')
        self.tree = et.Tree(f'({self._create_newick_string()})initial:0.0;', format=1)
        self._design_node_initial_state(tree=self.tree)
        # Add features/ faces to tree
        for index, key in enumerate(df_disp['DISPKEY']):
            # This should be unique get the escape nodes
            key_node = self.tree.search_nodes(name=key)[0]
            self._design_node_displaced_state(node=key_node, df_disp=df_disp, index=index, key=key)
            escape_nodes = key_node.get_children()
            # get sub data frame
            df_esc_filtered = df_esc[df_esc['DISPKEY'] == key]
            # iterate through escape nodes
            for escape_node in escape_nodes:
                # Get folder for these escape nodes
                self._design_node_escaped_state(node=escape_node, df_esc=df_esc_filtered)
                # go through converge stage
                conv_nodes = escape_node.get_children()
                df_conv_filtered = df_conv[df_conv['DISPKEY'] == key]
                df_conv_filtered = df_conv_filtered[df_conv_filtered['ESCKEY'] == escape_node.name]
                for conv_node in conv_nodes:
                    self._design_node_converged_state(node=conv_node, df_conv=df_conv_filtered)
                    nodes_at_current_depth = True
                    curr_node = conv_node
                    while nodes_at_current_depth:
                        try:
                            curr_node = curr_node.get_children()[0]
                            if curr_node.name in ['convex', 'sp']:
                                escapemodename = curr_node.up.name
                                df_conv_filtered = df_conv[df_conv['DISPKEY'] == key]
                                df_conv_filtered = df_conv_filtered[df_conv_filtered['ESCKEY'] == escapemodename]
                                self._design_node_reescape_conv_state(node=curr_node, df_conv=df_conv_filtered)
                            else:
                                df_esc_filtered = df_esc[df_esc['DISPKEY'] == key]
                                self._design_node_reescape_esc_state(node=curr_node, df_esc=df_esc_filtered)

                        except IndexError:
                            nodes_at_current_depth = False
                            print('reached leaf')
        self.ts = et.TreeStyle()
        # Do not add leaf names automatically
        self.ts.show_leaf_name = False
        self.ts.mode = 'c'
        self.ts.show_scale = False
        self.ts.show_branch_length = False
        self.ts.root_opening_factor = 0.0
        self.ts.arc_span = 360
        # Use my custom layout
        self.ts.layout_fn = my_layout

    def show_ete_tree(self):
        r"""
        Shows ete tree
        """
        self.tree.show(tree_style=self.ts)

    def render_ete_tree(self, filename:str = 'tree_ete.pdf'):
        r"""
        Args:
            filename: name and format of the file
        """
        self.tree.render(filename,tree_style=self.ts)

    def _create_tree_treelib(self):
        r"""
        Simple treelib representation of tree
        """
        l_tree = Tree()
        df_disp = pd.read_csv(self.info_file_disp, sep=r'\s+')
        df_esc = pd.read_csv(self.info_file_esc, sep=r'\s+')
        df_conv = pd.read_csv(self.info_file_conv, sep=r'\s+')
        # root node
        l_tree.create_node('initial configuration', 'root')
        for index, key in enumerate(df_disp['DISPKEY']):
            l_tree.create_node(df_disp['spin_file'][index], key, parent='root')
        for index, key in enumerate(df_esc['DISPKEY']):
            l_tree.create_node(df_esc['spin_file'][index], key + df_esc['ESCKEY'][index], parent=key)
        for index, key in enumerate(df_conv['DISPKEY']):
            l_tree.create_node(df_conv['spin_file'][index],
                               key + df_conv['ESCKEY'][index] + df_conv['spin_file'][index],
                               parent=key + df_conv['ESCKEY'][index])
            l_tree.create_node(df_conv['sp_or_convex'][index],
                               key + df_conv['ESCKEY'][index] + df_conv['sp_or_convex'][index],
                               parent=key + df_conv['ESCKEY'][index] + df_conv['spin_file'][index])
        l_tree.save2file(filename=self.parent_directory / 'sps_tree.txt')



r"""
Abstract base class for topology classes
"""
from abc import ABC, abstractmethod
import numpy as np


class ITopology(ABC):
    r"""
    abstract base class for inheriting topology classes for spin lattices
    """

    @property
    @abstractmethod
    def dual_lattice_points(self) -> np.ndarray:
        r"""
        Returns the points of the dual lattice
        """

    @property
    @abstractmethod
    def areas1(self) -> np.ndarray:
        r"""
        Returns the area of the upper triangles
        """

    @property
    @abstractmethod
    def areas2(self) -> np.ndarray:
        r"""
        Returns the area of the upper triangles
        """

    @property
    @abstractmethod
    def topo_dens(self) -> np.ndarray:
        r"""
        Returns the topological density for each dual lattice point
        """

    @property
    @abstractmethod
    def polygons1(self) -> np.ndarray:
        r"""
        Returns the upper triangles
        """

    @property
    @abstractmethod
    def polygons2(self) -> np.ndarray:
        r"""
        Returns the lower triangles
        """

    @property
    @abstractmethod
    def topological_center(self) -> np.ndarray:
        r"""
        Returns the center of mass with respect to the topological density
        """


# -*- coding: utf-8 -*-
r"""
Classes responsible for rotations of spin configurations
"""
import numpy as np
from abc import abstractmethod, ABC
from pathlib import Path
import pandas as pd
from typing import Union


class IRotation(ABC):
    r"""
    Abstract base class for rotations of spin configurations
    """

    def __init__(self, spins: np.ndarray, scale: float = 1.0, rodriguez_threshold: float = 1.0e-5) -> None:
        r"""
        Initializes the superclass for rotations

        Args:
            spins: spin array of all spins in the lattice: [[s1x,s1y,s1z],...,[sNx,sNy,sNz]]
            scale: scales the rotations angle for all spins
            rodriguez_threshold: defines how small an angle for a certain threshold has to be in order to switch the
            description of rodriguez formula to the corresponding taylor expansion
        """
        self._spins = spins
        self._scale = scale
        self._rodriguez_threshold = rodriguez_threshold

    @abstractmethod
    def applyrotation(self) -> np.ndarray:
        r"""
        Applies rotations
            returns:
                the rotated spin configuration: [[s1x',s1y',s1z'],...,[sNx',sNy',sNz']]
        """


class CRotationAlongEvec(IRotation):
    r"""
    Class for rotation of a spin lattice along an eigenvector.
    """

    def __init__(self, spins: np.ndarray, scale: float = 1.0, evec_source: str = 'file',
                 evec_path: Path = Path.cwd() / 'evec.dat', evec_data: Union[None, np.ndarray] = None) -> None:
        r"""
        Initializes the rotation of a spin lattice along an eigenvector.

        Args:
            spins: spin array of all spins in the lattice: [[s1x,s1y,s1z],...,[sNx,sNy,sNz]]
            scale: scales the rotations angle for all spins
            evec_source: decider string from which source the data for the eigenvector shall be read. Can be file
            (data is read from the file provided in 'evec_path') or data (the data is used provided in 'evec_data').
            evec_path: path to the eigenvector file
            evec_data: numpy array for the eigenvector information. Default is None here.
        """
        super().__init__(spins, scale=scale, rodriguez_threshold=1e-5)
        if evec_source == 'data':
            if evec_data is None:
                raise ValueError(
                    'If you choose data as as source please provide the eigenvector array in the evec_data keyword!')
            else:
                self._displacement = evec_data
        elif evec_source == 'file':
            try:
                self._df_evec = pd.read_csv(evec_path, sep=r'\s+', usecols=[3, 4, 5, 6],
                                            names=['vx', 'vy', 'vz', 'v'])
                self._displacement = np.zeros(shape=(len(self._df_evec['v']), 3))
                self._displacement[:, 0] = self._df_evec['vx'].to_numpy()
                self._displacement[:, 1] = self._df_evec['vy'].to_numpy()
                self._displacement[:, 2] = self._df_evec['vz'].to_numpy()
            except FileNotFoundError:
                raise FileNotFoundError(
                    'Corresponding eigenvector file not found. Please provide correct path under keyword: evec_path')
        else:
            raise ValueError('Keyword for eigenvector data source not supported!')
        # apply scaling:
        self._displacement = self._displacement * self._scale

    def applyrotation(self) -> np.ndarray:
        r"""
        Applies rotations
            returns:
             the rotated spin configuration: [[s1x',s1y',s1z'],...,[sNx',sNy',sNz']]
        """
        rotated_spins = np.zeros(shape=(len(self._df_evec['v']), 3))
        for (index, spin) in enumerate(self._spins):
            theta_i = np.linalg.norm(self._displacement[index, :])
            if theta_i >= self._rodriguez_threshold:
                rotated_spins[index, :] = spin * np.cos(theta_i) + self._displacement[index, :] * np.sin(
                    theta_i) / theta_i
            else:
                rotated_spins[index, :] = spin * np.cos(theta_i) + self._displacement[index, :] * (
                            1.0 - 1.0 / 6.0 * theta_i ** 2 + 1.0 / 120.0 * theta_i ** 4)
        return rotated_spins



# -*- coding: utf-8 -*-
r"""
Utilities for lattice operations
"""
import numpy as np
from scipy.optimize import fmin


def theta(r: np.ndarray, c: float, w: float) -> np.ndarray:
    r"""
    Theta function for example needed to create skyrmion
    Args:
        c(float): size of the domain in the middle of the skyrmion
        w(float): size of the region where the spins tilt (domain wall width)
    """
    comp1 = np.arcsin(np.tanh((-r - c) * 2 / w))
    comp2 = np.arcsin(np.tanh((-r + c) * 2 / w))
    return np.pi + comp1 + comp2


def dtheta_dr(r: np.ndarray, c: float, w: float) -> np.ndarray:
    r"""
    Calculates the derivative of the theta function of the skyrmion.

    Args:
        r(np.ndarray): 2-dimensional points of the lattice
        c(float): size of the domain in the middle of the skyrmion
        w(float): size of the region where the spins tilt (domain wall width)

    Returns:
        the derivative at the input points
    """
    comp1 = 2.0 * np.sqrt(-np.tanh(2.0 * (-c + np.abs(r)) / w) ** 2 + 1) / w
    comp2 = 2.0 * np.sqrt(-np.tanh(2.0 * (c + np.abs(r)) / w) ** 2 + 1) / w
    return -comp1 - comp2


def phi(p: np.ndarray, vorticity: float, helicity: float) -> np.ndarray:
    r"""
    Theta function for example needed to create skyrmion
    """
    return vorticity * p + helicity


def sk_2dprofile(r: np.ndarray, centerx: float, centery: float, c: float, w: float) -> np.ndarray:
    r"""
    Args:
        r(np.ndarray): 2-dimensional points of the lattice
        centerx(float): (2d) xcenter of the skyrmion
        centery(float): (2d) ycenter of the skyrmion
        c(float): size of the domain in the middle of the skyrmion
        w(float): size of the region where the spins tilt (domain wall width)
    Returns:
        the theta angle calculated with sz-components of the skyrmion profile.
    """
    center = np.array([centerx, centery])
    if (np.shape(r)[1] != 2) or (len(center) != 2):
        raise ValueError('Wrong dimension of input points or center variable.')
    return np.cos(theta(np.linalg.norm(r - center,axis=1), c, w))


def sk_radius(c: float, w: float) -> float:
    r"""
    Calculates the skyrmion radius using lilley criteria

    Args:
        c(float): size of the domain in the middle of the skyrmion
        w(float): size of the region where the spins tilt (domain wall width)
    Returns:
        the skyrmion radius in units of the lattice constant.
    """
    x0 = fmin(lambda x: dtheta_dr(x, c, w), 0, disp=False, xtol=1.0E-10)[0]
    rad = x0 - theta(x0, c, w) / dtheta_dr(x0, c, w)
    return rad


def spherical_area(s1: np.ndarray, s2: np.ndarray, s3: np.ndarray) -> float:
    r"""
    Calculates the spherical area of three spins. The sign of the area is the sign of the product:
    s1 * (s2 x s3).
    Implementation taken from userlibrary by Moritz Goerzen

    Args:
        s1: spin 1
        s2: spin 2
        s3: spin 3
    Returns:
        the spherical area
    """
    con = 1 + np.dot(s1, s2) + np.dot(s2, s3) + np.dot(s3, s1)
    # all spins are parallel -> no spherical area
    if con >= 4. - 10e-8:
        area = 0.
    else:
        mat = [s1, s2, s3]
        det_mat = np.linalg.det(mat)
        # EXCEPTIONAL CASE
        if abs(con) <= 10e-8 and abs(det_mat) <= 10e-8:
            area = 0.
        # NON-EXCEPTIONAL NON-COLINEAR CASE
        else:
            area = 2.0 * np.arctan2(det_mat, con)
    return area




# -*- coding: utf-8 -*-
r"""
Class for the hexagonal lattice. Stacked in the aa stacking order. All atoms sit on the same lattice sites.
"""
from povray.states_pyvista.ILattice import ILattice
from povray.states_pyvista.CCubicLattice import CCubicLatticeStackAA
import numpy as np
import povray.states_pyvista.utilities as utils
import pyvista as pv
from typing import List
import matplotlib


class CHexLatticeStackAA(ILattice):
    r"""
    Number of hexagonal lattices stacked directly above each other. This means that a3||e_z.
    """

    def __init__(self, ip_lattconst: float = 1, oop_lattconst: float = 1, N1: int = 50, N2: int = 50,
                 N3: int = 1, magdir: np.ndarray = np.array([0.0, 0.0, 1.0]), magmom: float = 1.0) -> None:
        r"""
        Initializes the lattice

        Args:
            ip_lattconst(float): in plane lattice constant
            oop_lattconst(float): out of plane lattice constant
            N1(int): number of unit cells in a1 dir.
            N2(int): number of unit cells in a2 dir.
            N3(int): number of unit cells in a2 dir.
            magdir(array): initial direction of the magnetization
            magmom(float): magnetic moment of lattice
        """
        a1 = np.array([0.5, np.sqrt(3) / 2, 0]) * ip_lattconst
        a2 = np.array([0.5, -np.sqrt(3) / 2, 0]) * ip_lattconst
        a3 = np.array([0, 0, 1]) * oop_lattconst
        super().__init__(a1=a1, a2=a2, a3=a3, N1=N1, N2=N2, N3=N3, magdir=magdir, magmom=magmom)

    def add_skyrmiontube(self, vorticity: int = 1.0, helicity: float = np.pi / 1, c: float = 2.5, w: float = 2.0,
                         AFM: bool = False) -> None:
        r"""
        Adds a skyrmion tube in all layers. Each layer has the same skyrmion (same parameters). Does also works for
        monolayer skyrmions

        Args:
            vorticity(int): topological charge of the skyrmion
            helicity(float): pi -> neel, pi/2 -> bloch
            c(float): size of the domain in the middle of the skyrmion
            w(float): size of the region where the spins tilt (domain wall width)
            AFM(bool): whether each layer shall be antiferromagnet
        """
        for layer in range(self.nlayer):
            magstructure_layer = self.getlayer_by_idx(layer)
            XY = magstructure_layer[:, :2].copy() - self.layermidpoints[layer][:2]
            r = np.linalg.norm(XY, axis=1)
            pp = np.arctan2(XY[:, 1], XY[:, 0])
            th = utils.theta(r, c, w)
            ph = utils.phi(pp, vorticity, helicity)
            n = np.arange(0, len(XY[:, 0]), 1)
            if AFM:
                sign = (-1) ** (n % 2 + n // int(np.sqrt(len(XY[:, 0]))))
            else:
                sign = 1
            uplo = 1
            magstructure_layer[:, 3] = np.sin(th) * np.cos(ph) * sign
            magstructure_layer[:, 4] = np.sin(th) * np.sin(ph) * sign
            magstructure_layer[:, 5] = np.cos(th) * float(uplo) * sign
            self.setlayer_by_idx(layer, magstructure_layer)

    def add_SAF_skyrmiontube(self, vorticity: int = 1.0, helicity: float = np.pi / 1, c: float = 2.5, w: float = 3.0,
                             AFM: bool = False) -> None:
        r"""
        Adds a synthethic antiferromagnetic skyrmion tube in all layers. Each layer has the same skyrmion
        (same parameters). Does also works for monolayer skyrmions

        Args:
            vorticity(int): topological charge of the skyrmion
            helicity(float): pi -> neel, pi/2 -> bloch
            c(float): size of the domain in the middle of the skyrmion
            w(float): size of the region where the spins tilt (domain wall width)
            AFM(bool): whether each layer shall be antiferromagnet
        """
        for layer in range(self.nlayer):
            magstructure_layer = self.getlayer_by_idx(layer)
            XY = magstructure_layer[:, :2].copy() - self.layermidpoints[layer][:2]
            r = np.linalg.norm(XY, axis=1)
            pp = np.arctan2(XY[:, 1], XY[:, 0])
            th = utils.theta(r, c, w)
            ph = utils.phi(pp, vorticity, helicity)
            n = np.arange(0, len(XY[:, 0]), 1)
            if AFM:
                sign = (-1) ** (n % 2 + n // int(np.sqrt(len(XY[:, 0]))))
            else:
                sign = 1
            uplo = 1
            magstructure_layer[:, 3] = np.sin(th) * np.cos(ph) * sign
            magstructure_layer[:, 4] = np.sin(th) * np.sin(ph) * sign
            magstructure_layer[:, 5] = np.cos(th) * float(uplo) * sign
            if layer % 2 == 0:
                magstructure_layer[:, 3:6] = magstructure_layer[:, 3:6] * -1
            self.setlayer_by_idx(layer, magstructure_layer)

    def add_chiralbobber(self, vorticity: int = 1.0, helicity: float = np.pi / 1, c: List[float] = [0.0, 1.0, 1.5, 2.5],
                         w: List[float] = [0.0, 1.5, 2.0, 3.0],
                         AFM: bool = False) -> None:
        r"""
        Adds a skyrmion tube in all layers. Each layer has the same skyrmion (same parameters). Does also works for
        monolayer skyrmions

        Args:
            vorticity(int): topological charge of the skyrmion
            helicity(float): pi -> neel, pi/2 -> bloch
            c(float): size of the domain in the middle of the skyrmion
            w(float): size of the region where the spins tilt (domain wall width)
            AFM(bool): whether each layer shall be antiferromagnet
        """
        for layer in range(self.nlayer):
            magstructure_layer = self.getlayer_by_idx(layer)
            if layer >= len(c) or c[layer] == 0.0:
                continue
            XY = magstructure_layer[:, :2].copy() - self.layermidpoints[layer][:2]
            r = np.linalg.norm(XY, axis=1)
            pp = np.arctan2(XY[:, 1], XY[:, 0])
            th = utils.theta(r, c[layer], w[layer])
            ph = utils.phi(pp, vorticity, helicity)
            n = np.arange(0, len(XY[:, 0]), 1)
            if AFM:
                sign = (-1) ** (n % 2 + n // int(np.sqrt(len(XY[:, 0]))))
            else:
                sign = 1
            uplo = 1
            magstructure_layer[:, 3] = np.sin(th) * np.cos(ph) * sign
            magstructure_layer[:, 4] = np.sin(th) * np.sin(ph) * sign
            magstructure_layer[:, 5] = np.cos(th) * float(uplo) * sign
            self.setlayer_by_idx(layer, magstructure_layer)



# -*- coding: utf-8 -*-
r"""
Module contains abstract base class for visualizing spin structures / lattices.
"""
from abc import ABC, abstractmethod
from spinterface.inputs.lattice.ILattice import ILattice
from pathlib import Path


class IVisualizer(ABC):
    r"""
    Abstract class for visualizations of spin structures
    """

    def __init__(self, lattice: ILattice) -> None:
        r"""
        Initialization method for the visualizer. Will be called from sub class

        Args:
            lattice(ILattice): realization of ILattice
        """
        self.lattice = lattice

    @abstractmethod
    def __call__(self, outpath: Path = Path.cwd() / 'lattice.png'):
        r"""
        Call method for the visualization

        Args:
            outpath(Path): path for the output file
        """



r"""
Visualizer for the esc-stage of the sps algorithm
"""
from spinterface.sps_results.stage_visualizer.ivisstage import IVisStage
from pathlib import Path
from typing import Union, Tuple, List
from spinterface.inputs.lattice.CLattice import CLattice
from spinterface.visualizations.lattices.cvisualpyvista import CVisualPyVista
import pandas as pd


class CVisEsc(IVisStage):
    r"""

    """

    def __call__(self) -> None:
        r"""
        Excecutes the visualization for the displace stage
        """
        if self.strategy in [1]:
            self._visualize_escaped_configuration()
            self._visualize_eigenvector_start_end()

    def __init__(self, strategy: int, info_file: Union[Path, str],
                 camera_spin_files: Union[List[Tuple[float, float, float]], None] = None,
                 spinvisualization_technique: str = 'oop', arrowscale_evec: float = 30)->None:
        r"""
        Initializes the visualizer for the escape-stage of SPS
        Args:
            strategy(int): escape strategy
            info_file(Path,str): the file which holds information about the locations
            of the corresponding files. It is always in the same directory as the sps.in therefore all paths in the info
            file are relative to this.
        """
        super().__init__(strategy, camera_spin_files)
        self.parent_directory = Path(info_file).parent
        self.arrowscale_evec = arrowscale_evec
        self.keys, self.esckeys, self.folders, self.spinfiles = self._read_info_file(info_file=info_file)
        self.spinvisualization_technique = spinvisualization_technique

    @staticmethod
    def _read_info_file(info_file: Union[Path, str]) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series]:
        r"""
        Reads the columns of the escape-info file.
        """
        df = pd.read_csv(info_file, sep=r'\s+')
        return df['DISPKEY'], df['ESCKEY'], df['folder'], df['spin_file']

    @property
    def stage(self) -> str:
        r"""
        Returns:
            the stage
        """
        return 'Escape'

    def _stagestrategytext(self) -> str:
        r"""
        Returns:
            the escape strategy in words
        """
        if self.strategy == 1:
            return 'Use modefollowing to escape the convex region.'
        else:
            return 'Describing text for this strategy not yet coded.'

    def _visualize_escaped_configuration(self):
        r"""
        Visualizes the spin configurations for all escaped configurations
        """
        for spinfile in self.spinfiles:
            print(f'Visualizig: {spinfile}...')
            l_latt = CLattice(source='STM', path=self.parent_directory / spinfile)
            if self.spinvisualization_technique == 'oop':
                l_visu = CVisualPyVista(lattice=l_latt, cam=self.spincam)
            elif self.spinvisualization_technique == 'ipoop':
                l_visu = CVisualPyVista(lattice=l_latt, cam=self.spincam, heatmap=True, heatmap_saturation=1.0)
            else:
                raise ValueError('Selected spin visualization technique is not coded yet!')
            pngfile = spinfile.replace('.dat', '.png')
            pngfile_splitted = pngfile.split('/')
            pngpath = self.parent_directory
            for pathpart in pngfile_splitted:
                pngpath = pngpath / pathpart
            l_visu(outpath=pngpath)
            l_visu.plotter.close()
            print('...done')

    def _visualize_eigenvector_start_end(self):
        r"""
        Visualizes the spin configurations for all displaced configurations
        """
        for spinfile in self.spinfiles:
            evecfile_end = spinfile.replace('spin', 'evec')
            if (self.parent_directory / evecfile_end).is_file():
                print(f'Visualizig: {evecfile_end}...')
                l_latt_end = CLattice(source='evec', path=self.parent_directory / evecfile_end)
                l_visu_end = CVisualPyVista(lattice=l_latt_end, cam=self.spincam, arrowscale=self.arrowscale_evec)
                pngfile_end = evecfile_end.replace('.dat', '.png')
                pngfile_splitted_end = pngfile_end.split('/')
                pngpath_end = self.parent_directory
                for pathpart in pngfile_splitted_end:
                    pngpath_end = pngpath_end / pathpart
                l_visu_end(outpath=pngpath_end)
                l_visu_end.plotter.close()
                print('...done')
            evecfile_start = evecfile_end.replace('end', '0')
            if (self.parent_directory / evecfile_start).is_file():
                print(f'Visualizig: {evecfile_start}...')
                l_latt_start = CLattice(source='evec', path=self.parent_directory / evecfile_start)
                l_visu_start = CVisualPyVista(lattice=l_latt_start, cam=self.spincam, arrowscale=self.arrowscale_evec)
                pngfile_start = evecfile_start.replace('.dat', '.png')
                pngfile_splitted_start = pngfile_start.split('/')
                pngpath_start = self.parent_directory
                for pathpart in pngfile_splitted_start:
                    pngpath_start = pngpath_start / pathpart
                l_visu_start(outpath=pngpath_start)
                l_visu_start.plotter.close()
                print('...done')



r"""
Visualizer for the disp-stage of the sps algorithm
"""
from spinterface.sps_results.stage_visualizer.ivisstage import IVisStage
from pathlib import Path
from typing import Union, Tuple, List
from spinterface.inputs.lattice.CLattice import CLattice
from spinterface.visualizations.lattices.cvisualpyvista import CVisualPyVista
import pandas as pd


class CVisDisp(IVisStage):
    r"""

    """

    def __call__(self) -> None:
        r"""
        Excecutes the visualization for the displace stage
        """
        if self.strategy in [2]:
            self._visualize_displaced_configuration()
            self._visualize_superposition_eigenvector()

    def __init__(self, strategy: int, info_file: Union[Path, str],
                 camera_spin_files: Union[List[Tuple[float, float, float]], None] = None,
                 spinvisualization_technique: str = 'oop') -> None:
        r"""
        Initializes the visualizer for the displace-stage of SPS
        Args:
            strategy(int): displace strategy
            info_file(Path,str): the file which holds information about the locations
            of the corresponding files. It is always in the same directory as the sps.in therefore all paths in the info
            file are relative to this.
        """
        super().__init__(strategy, camera_spin_files)
        self.parent_directory = Path(info_file).parent
        self.keys, self.folders, self.spinfiles = self._read_info_file(info_file=info_file)
        self.spinvisualization_technique = spinvisualization_technique

    @staticmethod
    def _read_info_file(info_file: Union[Path, str]) -> Tuple[pd.Series, pd.Series, pd.Series]:
        r"""
        Reads the columns of the displace-info file.
        """
        df = pd.read_csv(info_file, sep=r'\s+')
        return df['DISPKEY'], df['folder'], df['spin_file']

    @property
    def stage(self) -> str:
        r"""
        Returns:
            the stage
        """
        return 'Displace'

    def _stagestrategytext(self) -> str:
        r"""
        Returns:
            the displace strategy in words
        """
        if self.strategy == 1:
            return 'Not displacing at all.'
        elif self.strategy == 2:
            return 'Creating a random superposition of eigenvectors of initial state and rotate conf. along this ' \
                   'superpos. '
        else:
            return 'Describing text for this strategy not yet coded.'

    def _visualize_displaced_configuration(self):
        r"""
        Visualizes the spin configurations for all displaced configurations
        """
        for spinfile in self.spinfiles:
            l_latt = CLattice(source='STM', path=self.parent_directory / spinfile)
            if self.spinvisualization_technique == 'oop':
                l_visu = CVisualPyVista(lattice=l_latt, cam=self.spincam)
            elif self.spinvisualization_technique == 'ipoop':
                l_visu = CVisualPyVista(lattice=l_latt, cam=self.spincam, heatmap=True, heatmap_saturation=1.0)
            else:
                raise ValueError('Selected spin visualization technique is not coded yet!')
            pngfile = spinfile.replace('.dat', '.png')
            pngfile_splitted = pngfile.split('/')
            pngpath = self.parent_directory
            for pathpart in pngfile_splitted:
                pngpath = pngpath / pathpart
            l_visu(outpath=pngpath)
            l_visu.plotter.close()

    def _visualize_superposition_eigenvector(self):
        r"""
        Visualizes the spin configurations for all displaced configurations
        """
        for spinfile in self.spinfiles:
            evecfile = spinfile.replace('spin', 'evec')
            l_latt = CLattice(source='evec', path=self.parent_directory / evecfile)
            l_visu = CVisualPyVista(lattice=l_latt, cam=self.spincam)
            pngfile = evecfile.replace('.dat', '.png')
            pngfile_splitted = pngfile.split('/')
            pngpath = self.parent_directory
            for pathpart in pngfile_splitted:
                pngpath = pngpath / pathpart
            l_visu(outpath=pngpath)
            l_visu.plotter.close()



from pathlib import Path
from spinterface.inputs.lattice.ILattice import ILattice
from spinterface.visualizations.lattices.ivisualizer import IVisualizer
from spinterface.visualizations.lattices.utilities import get_colormap
from typing import List, Union, Tuple
import pyvista as pv
import numpy as np


class CVisualPyvistaBilayer(IVisualizer):
    r"""
        It visualizes a bilayer spin lattice with three different perspectives. A
        view of the top layer, of the bot layer and a view perspective from the middle
    """

    def __init__(self, lattice: ILattice, tiplength: float = 1.0, tipradius: float = 0.35,
                 arrowscale: float = 0.7, cam_mid: Union[List[Tuple[float, float, float]], None] = None,
                 cmap: str = 'hsv_spind', dist_top: float = 10, dist_bot: float = 10,
                 filter_around_layercenter: Union[None, float] = None,
                 dist_betweenlayers: float = 3, filter_half: bool = False) -> None:
        r"""
        Initializes the visualization of a bilayer spin lattice.

        Args:
            tiplength(float): geometry of arrow: tiplength
            tipradius(float): geometry of arrow: tipradius
            arrowscale(float): geometry of arrow: arrowscale
            cam_mid: camera position for the middle layer
            dist_bot (float): distance of the camera (in z direction) from the bottom layer
            dist_top (float): distance of the camera (in z direction) from the top layer
            cmap: string for the choice of the colormap. Defined in utilities module
            filter_around_layercenter(float, None): if not none each layer is filtered according to a circle around
            its layer midpoint
            dist_betweenlayers(float): The distance between the two layers in the middle picture
            filter_half(bool): if the half of the cylinder should be displayer
        """
        if lattice.nlayer != 2:
            raise ValueError('This visualization class is only valid for bilayer systems.')
        super().__init__(lattice)
        self._geom = pv.Arrow(start=np.array([-arrowscale / 2.0, 0, 0]), tip_length=tiplength,
                              tip_radius=tipradius, scale=arrowscale)
        self.cmap = get_colormap(cmap)
        self.cam_mid = cam_mid
        self.dist_top = dist_top
        self.dist_bot = dist_bot
        self.dist_betweenlay = dist_betweenlayers
        self.filter_around_layercenter = filter_around_layercenter
        self.filter_half = filter_half
        self._make_plotter()
        if self.filter_around_layercenter is not None:
            self._make_cylinder()

    def _make_cylinder(self):
        r"""
        Adds a grey cylinder to the central image view from the side
        """
        l_cyl = pv.Cylinder(center=self.lattice.midpoint + np.array([0, 0, self.dist_betweenlay / 2]) - np.array(
            [0, 0, self.lattice.midpoint[2]]), direction=np.array([0, 0, 1]), radius=self.filter_around_layercenter,
                            height=self.dist_betweenlay + 2 * self.lattice.midpoint[2])
        l_cyl = l_cyl.clip(normal='y')
        self.plotter.subplot(0)
        self.plotter.add_mesh(l_cyl, opacity=1, color='gray')

    def _make_plotter(self, offscreen: bool = False):
        r"""
        Creates the plotter. The plotter will be recreated when saving the image
        """
        self.plotter = pv.Plotter(shape=('1|2'), off_screen=offscreen, lighting='three lights')
        self._configureplotter()
        plotpoints1, plotspins1, plotsz1 = self._make_plot_points(layer=0)
        plotpoints2, plotspins2, plotsz2 = self._make_plot_points(layer=1)
        # shift the upper layer
        plotpoints1_mid, plotspins1_mid, plotsz1_mid = self._make_plot_points(layer=0, filter_half=self.filter_half)
        plotpoints2_mid, plotspins2_mid, plotsz2_mid = self._make_plot_points(layer=1, filter_half=self.filter_half)
        plotpoints2_mid = plotpoints2_mid + np.array([0.0, 0.0, self.dist_betweenlay])

        self.PolyData1 = pv.PolyData(plotpoints1)
        self.PolyData1.vectors = plotspins1
        self.PolyData1['oop'] = plotsz1
        self.Glyphs1 = self.PolyData1.glyph(orient=True, scale=True, geom=self._geom)
        self.PolyData2 = pv.PolyData(plotpoints2)
        self.PolyData2.vectors = plotspins2
        self.PolyData2['oop'] = plotsz2
        self.Glyphs2 = self.PolyData2.glyph(orient=True, scale=True, geom=self._geom)

        self.PolyData2_mid = pv.PolyData(plotpoints2_mid)
        self.PolyData2_mid.vectors = plotspins2_mid
        self.PolyData2_mid['oop'] = plotsz2_mid
        self.Glyphs2_mid = self.PolyData2_mid.glyph(orient=True, scale=True, geom=self._geom)
        self.PolyData1_mid = pv.PolyData(plotpoints1_mid)
        self.PolyData1_mid.vectors = plotspins1_mid
        self.PolyData1_mid['oop'] = plotsz1_mid
        self.Glyphs1_mid = self.PolyData1_mid.glyph(orient=True, scale=True, geom=self._geom)

        self.plotter.subplot(1)
        self.plotter.add_mesh(self.Glyphs2, show_scalar_bar=False, cmap=self.cmap)
        self.plotter.subplot(0)
        self.plotter.add_mesh(self.Glyphs1_mid, show_scalar_bar=False, cmap=self.cmap)
        self.plotter.add_mesh(self.Glyphs2_mid, show_scalar_bar=False, cmap=self.cmap)
        self.plotter.subplot(2)
        self.plotter.add_mesh(self.Glyphs1, show_scalar_bar=False, cmap=self.cmap)

    def _make_plot_points(self, layer: int = 1, filter_half: bool = False) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        r"""
        We always want to norm the colormap in the interval -1, 1 even we have a lattice which spins have only SZ comp.
        in the interval e.g. (1,0.5). There is now easy way to do this with pyvista since there is no interface for nor-
        malizing. Therefore, we add an invisible point in the center of the lattice here.

        Returns:
            the points, the spins and the sz components
        """
        if self.filter_around_layercenter is not None:
            mag_structure_layer = self._filter_layercenter(magstructure=self.lattice.getlayer_by_idx(layer),
                                                           layer=layer)
        else:
            mag_structure_layer = self.lattice.getlayer_by_idx(layer)
        if filter_half:
            mag_structure_layer = mag_structure_layer[mag_structure_layer[:, 1] <= 0]
        points = mag_structure_layer[:, :3]
        spins = mag_structure_layer[:, 3:6]
        SZ = mag_structure_layer[:, 5]
        midpoint = self.lattice.layermidpoints[layer]
        plotpoints = np.append(points, np.array([midpoint, midpoint]), axis=0)
        plotspins = np.append(spins, np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]), axis=0)
        plotsz = np.append(SZ, np.array([1.0, -1.0]))
        return plotpoints, plotspins, plotsz

    def _filter_layercenter(self, magstructure: np.ndarray, layer: int = 0) -> np.ndarray:
        r"""
        Applies a circular filter to a layer.

        Args:
            layer(int): adress of the layer
            magstructure(np.ndarray): struct. to filter
        Returns:
            updated magstructure
        """
        magstructure_layer = self.lattice.getlayer_by_idx(layer)
        mp = self.lattice.layermidpoints[layer]
        new_structure_layer = magstructure_layer[
            np.linalg.norm(magstructure_layer[:, :3] - mp, axis=1) <= self.filter_around_layercenter]
        return new_structure_layer

    def _configureplotter(self) -> None:
        r"""
        Configures the plotter object
        """
        pv.set_plot_theme("ParaView")
        pv.rcParams['transparent_background'] = True
        self.plotter.subplot(1)
        self.plotter.set_background('white')
        mp = self.lattice.layermidpoints[1]
        self.plotter.camera_position = [(mp[0], mp[1], mp[2] + self.dist_top), (mp[0], mp[1], mp[2]), (0, 1, 0)]
        self.plotter.subplot(0)
        self.plotter.set_background('white')
        self.plotter.camera_position = self.cam_mid
        self.plotter.subplot(2)
        mp = self.lattice.layermidpoints[0]
        self.plotter.camera_position = [(mp[0], mp[1], mp[2] + self.dist_bot), (mp[0], mp[1], mp[2]), (0, 1, 0)]
        self.plotter.set_background('white')

        def camtop() -> None:
            self.plotter.subplot(1)
            print('Camera postion: ', self.plotter.camera_position)

        def cambot() -> None:
            self.plotter.subplot(2)
            print('Camera postion: ', self.plotter.camera_position)

        def cammid() -> None:
            self.plotter.subplot(0)
            print('Camera postion: ', self.plotter.camera_position)

        def help() -> None:
            print('the commands and keys are: camtop: t, cambot: b, cammid: c')

        self.plotter.add_key_event('h', help)
        self.plotter.add_key_event('c', cammid)
        self.plotter.add_key_event('t', camtop)
        self.plotter.add_key_event('b', cambot)

    def __call__(self, outpath: Path = Path.cwd() / 'bilayerspin.png') -> None:
        r"""
        Saves the image to a file

        Args:
            outpath(Path): output path for the png image created.
        """
        self._make_plotter(offscreen=True)
        if self.filter_around_layercenter:
            self._make_cylinder()
        self.plotter.window_size = [3000, 2000]
        self.plotter.screenshot(str(outpath.stem))

    def show(self) -> None:
        r"""
        Shows the plotter
        """
        print('Look what you have done.......')
        print('to get current cam-position press key c')
        self.plotter.show()



# -*- coding: utf-8 -*-
r"""
This module contains the class CLattice. Its realizations represent a spin structure.
"""
import pandas as pd
import spinterface.inputs.lattice.utilities as utils
from spinterface.inputs.lattice.ILattice import ILattice
from spinterface.inputs.lattice.spintransformations.CRotation import CRotationAlongEvec
from pathlib import Path
import numpy as np
from typing import Tuple, List, Union
from scipy.optimize import curve_fit
from spinterface.inputs.lattice.const import LATT_TYPE_SPIN, LATT_TYPE_EVEC, LATT_TYPE_FORCE
from spinterface.inputs.lattice.topology.CTopology import CTopologyHexCubic
from spinterface.inputs.lattice.topology.ITopology import ITopology
from spinterface.inputs.lattice.modeconstruction.cconstructiontranslational import CConstructTranslational


class CLattice(ILattice):
    r"""
    Creates a lattice which can be used to produce a SpinSTMi-type file. The structure will be read through a lattice.in
    file.
    """

    def __init__(self, source: str = 'lattice.in', path: Path = Path.cwd() / 'lattice.in',
                 magdir: np.array = np.array([0.0, 0.0, 1.0]), N1: Union[None, int] = None, N2: Union[None, int] = None,
                 N3: Union[None, int] = None) -> None:
        r"""
        Initializes the lattice

        Args:
            source(str): Defines how the lattice shall be constructed. If source is evec the spins will be the components
            of the eigenvector and the magmoms the length of these components.
            latticefile(Path): path to the lattice.in/STM/evec file from which the lattice will be constructed.
            magdir(np.array): initial magnetisation direction of the lattice.
            N1, N2, N3: in case of reading from STM file one might need the information about N1, N2, and N3 to calculate
            e.g. the topological density.
        """
        self._source = source
        if self._source == 'lattice.in':
            self._latticefile = path
            a1, a2, a3, r_motif, N1, N2, N3 = self._readfromlatticefile()
            super().__init__(a1, a2, a3, r_motif, N1, N2, N3, magdir)
        elif self._source in ['STM','evec','force']:
            self._stmfile = path
            a1, a2, a3, r_motif = None, None, None, None
            super().__init__(a1, a2, a3, r_motif, N1, N2, N3, magdir)
            df = pd.read_csv(path, sep=r'\s+', usecols=[0, 1, 2, 3, 4, 5, 6],
                             names=['x', 'y', 'z', 'sx', 'sy', 'sz', 'm'])
            self._points = np.column_stack((df['x'].to_numpy(), df['y'].to_numpy(), df['z'].to_numpy()))
            self._spins = np.column_stack((df['sx'].to_numpy(), df['sy'].to_numpy(), df['sz'].to_numpy()))
            self._magmoms = df['m'].to_numpy()
        else:
            raise NotImplementedError(f'Source not yet implemented. Choose between lattice.in or STM, evec or force.')

        self._skradius = None
        self._topologies = []

    def _readfromlatticefile(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[np.array], int, int, int]:
        r"""
        Reads the lattice.in.

        Returns:
            a1(np.array): lattice vector a1, multiplied by length stored in alat
            a2(np.array): lattice vector a2, multiplied by length stored in alat
            a3(np.array): lattice vector a3, multiplied by length stored in alat
            r_motif(List(np.array)): positions of the atoms within the unit cell, the corresponding vectors are gien in
            units of a1,a2 and a3.
            N1(int): number of unit cells in a1 direction
            N2(int): number of unit cells in a2 direction
            N3(int): number of unit cells in a3 direction
        """
        with open(str(self._latticefile), 'r') as f:
            for (lnr, line) in enumerate(f):
                L = str(line).lstrip()
                if L.startswith('#'):
                    continue
                if L.startswith('Nsize'):
                    L = L[5:]
                    L.lstrip()
                    ns = L.split()
                    N1 = int(ns[0])
                    N2 = int(ns[1])
                    N3 = int(ns[2])
                if L.startswith('alat'):
                    L = L[4:]
                    L.lstrip()
                    alats = L.split()
                    alat1 = int(float(alats[0]))
                    alat2 = int(float(alats[1]))
                    alat3 = int(float(alats[2]))
                if L.startswith('lattice'):
                    lattline = lnr
                if L.startswith('motif'):
                    L = L[5:]
                    motifline = lnr
                    Nuc = int(L.split()[0])

        with open(str(self._latticefile), 'r') as f:
            for (lnr, line) in enumerate(f):
                L = str(line)
                if lnr == lattline + 1:
                    L.lstrip()
                    a1s = L.split()
                    a1 = np.array([float(a1s[0]), float(a1s[1]), float(a1s[2])])
                if lnr == lattline + 2:
                    L.lstrip()
                    a2s = L.split()
                    a2 = np.array([float(a2s[0]), float(a2s[1]), float(a2s[2])])
                if lnr == lattline + 3:
                    L.lstrip()
                    a3s = L.split()
                    a3 = np.array([float(a3s[0]), float(a3s[1]), float(a3s[2])])
        r_motif = []
        for n in range(Nuc):
            with open(str(self._latticefile), 'r') as f:
                for (lnr, line) in enumerate(f):
                    L = str(line)
                    if lnr == motifline + 1 + n:
                        L.lstrip()
                        r1s = L.split()
                        r_motif.append(np.array([float(r1s[0]), float(r1s[1]), float(r1s[2]), float(r1s[3])]))
        # remove non magnetic atoms from unit cell
        r_motif = [mot for mot in r_motif if mot[3] != 0.0]

        return a1 * alat1, a2 * alat2, a3 * alat3, r_motif, N1, N2, N3

    def add_domainwallstarter(self, boundary_normal: np.ndarray = None, turn_boundary: bool = True,
                              width_boundary_region: float = 1.0) -> None:
        r"""
        Adds a heavyside function to a lattice. This means the spins on one side of the lattice point upwards while the
        other side points downwards. There is one line of spins in plane at the border. The normal direction to the boundary
        can be defined.

        Args:
            boundary_normal: vector defining the normal of the boundary.
            turn_boundary: weither the boundary line of spins shall lie in the in plane direction
            width_boundary_region: allowance for spins to be in the boundary. C
        """

        if self.source == LATT_TYPE_EVEC:
            raise ValueError('Adding a domain wall initial state to an eigenvector lattice makes no sense!')
        if self.source == LATT_TYPE_FORCE:
            raise ValueError('Adding a domain wall initial state to an force lattice makes no sense!')
        # make sure boundary vector is normalized
        if boundary_normal is None:
            boundary_normal = self.a1
        if boundary_normal[2] != 0:
            raise ValueError('No out of plane pointing boundaries are allowed.')
        boundary_normal = boundary_normal / np.linalg.norm(boundary_normal)
        for layer in range(self.nlayer):
            magstructure_layer = self.getlayer_by_idx(layer)
            midpoint = self.layermidpoints[layer]
            points = magstructure_layer[:, :3]
            dots = [np.dot((p - midpoint) / np.linalg.norm(p - midpoint), boundary_normal) for p in points]
            spins = [np.array([0.0, 0.0, 1.0]) if d >= 0 else np.array([0.0, 0.0, -1.0]) for d in dots]
            if turn_boundary:
                midpoint_shifted_forward = midpoint + boundary_normal * width_boundary_region
                midpoint_shifted_backward = midpoint - boundary_normal * width_boundary_region
                dots_forward = [np.dot((p - midpoint_shifted_forward) / np.linalg.norm(p - midpoint_shifted_forward),
                                       boundary_normal)
                                for p in points]
                dots_backward = [np.dot((p - midpoint_shifted_backward) / np.linalg.norm(p - midpoint_shifted_backward),
                                        boundary_normal)
                                 for p in points]

                spins = [boundary_normal if (0 < dots_backward[index]) and (d <= 0) else spins[index] for (index, d) in
                         enumerate(dots_forward)]
            magstructure_layer[:, 3:6] = np.asarray(spins)
            self.setlayer_by_idx(layer, magstructure_layer)

    def add_skyrmiontube(self, vorticity: int = 1.0, helicity: float = np.pi / 1, c: float = 2.5, w: float = 2.0,
                         AFM: bool = False) -> None:
        r"""
        Adds a skyrmion tube in all layers. Each layer has the same skyrmion (same parameters). Does also works for
        monolayer skyrmions

        Args:
            vorticity(int): topological charge of the skyrmion
            helicity(float): pi -> neel, pi/2 -> bloch
            c(float): size of the domain in the middle of the skyrmion
            w(float): size of the region where the spins tilt (domain wall width)
            AFM(bool): whether each layer shall be antiferromagnet
        """
        if self.source == LATT_TYPE_EVEC:
            raise ValueError('Adding a skyrmiontube to an eigenvector lattice makes no sense!')
        if self.source == LATT_TYPE_FORCE:
            raise ValueError('Adding a skyrmiontube to an force lattice makes no sense!')
        for layer in range(self.nlayer):
            magstructure_layer = self.getlayer_by_idx(layer)
            XY = magstructure_layer[:, :2].copy() - self.layermidpoints[layer][:2]
            r = np.linalg.norm(XY, axis=1)
            pp = np.arctan2(XY[:, 1], XY[:, 0])
            th = utils.theta(r, c, w)
            ph = utils.phi(pp, vorticity, helicity)
            n = np.arange(0, len(XY[:, 0]), 1)
            if AFM:
                sign = (-1) ** (n % 2 + n // int(np.sqrt(len(XY[:, 0]))))
            else:
                sign = 1
            uplo = 1
            magstructure_layer[:, 3] = np.sin(th) * np.cos(ph) * sign
            magstructure_layer[:, 4] = np.sin(th) * np.sin(ph) * sign
            magstructure_layer[:, 5] = np.cos(th) * float(uplo) * sign
            self.setlayer_by_idx(layer, magstructure_layer)

    def construct_translational_mode(self, lattvec: Union[np.ndarray, None] = None, write_file: bool = False,
                                     outpath: Path = Path.cwd() / 'evec_translation.dat') -> np.ndarray:
        r"""

        """
        if lattvec is None:
            if self.a1 is None or self.a2 is None:
                raise ValueError("Need lattice vector. Either provide as input to this routine or via lattice.in")
            l_lattvec = self.a1
        else:
            l_lattvec = lattvec
        evecconstructor = CConstructTranslational(spins=self.spins, points=self.points, lattvec=l_lattvec)
        evec = evecconstructor()
        if write_file:
            data = {'x': self.points[:, 0], 'y': self.points[:, 1], 'z': self.points[:, 2], 'vx': evec[:, 0],
                    'vy': evec[:, 1], 'vz': evec[:, 2], 'v': np.linalg.norm(evec,axis=1)}
            df = pd.DataFrame(data)
            df.to_csv(outpath,header=False,sep=' ',index=False)
        return evec

    def construct_rotational_mode(self) -> None:
        r"""

        :return:
        """
        raise ValueError('not yet coded!')

    def rotate(self, displacement: Union[Path, np.ndarray], scale: float = 1.0) -> None:
        r"""
        Rotates the spinconfiguration of the lattice along the displacement according to Rodrigues formula.
        :param displacement: 2D np array: [[p1x,p1y,p1z],...,[pNx,pNy,pNz]] where N is the number of spins in the lattice.
            The rotation axis for the ith spin is given by s_i x p_i and its rotation angle is defined as |p_i|.
        :param scale: Scales the rotation angle of all spins
        """
        if self.source == LATT_TYPE_EVEC:
            raise ValueError('Rotating the eigenvector itself does not make sense!')
        if self.source == LATT_TYPE_FORCE:
            raise ValueError('Rotating the force does not make sense!')
        l_spins = self.spins
        # decide on the source of the rotations:
        if type(displacement) == type(Path.cwd()):
            rotator = CRotationAlongEvec(spins=l_spins, scale=scale, evec_source='file', evec_path=displacement)
        elif type(displacement) == type(np.array([0.0, 0.0, 0.0])):
            rotator = CRotationAlongEvec(spins=l_spins, scale=scale, evec_source='data', evec_data=displacement)
        else:
            raise ValueError('Not the correct data type for displacement input.')
        self.spins = rotator.applyrotation()

    @property
    def source(self) -> str:
        r""""
        Returns:
            the source of the lattice. For lattice.in and STM a spin configuration is assumed. For
        """
        if self._source == 'lattice.in' or self._source == 'STM':
            return LATT_TYPE_SPIN
        elif self._source == 'evec':
            return LATT_TYPE_EVEC
        elif self._source == 'force':
            return LATT_TYPE_FORCE

    @property
    def skradius(self) -> List[Union[None, float]]:
        r"""
        Returns:
             the skyrmion radius in units of the lattice constant for all layers. None if no skyrmion exists or the fit
             fails.
        """
        if self.source == LATT_TYPE_EVEC:
            raise ValueError('Calculating the skyrmion radius for an eigenvector lattice makes no sense!')
        if self.source == LATT_TYPE_FORCE:
            raise ValueError('Calculating the skyrmion radius for a force lattice makes no sense!')
        rads = []
        for n in range(self.nlayer):
            currentlayer = self.getlayer_by_idx(n)
            if np.min(currentlayer[:, 3:6]) >= 0.0:
                print(f'layer {n} does not contain skyrmion.')
                rads.append(None)
            else:
                minmag = currentlayer[currentlayer[:, 5] == np.min(currentlayer[:, 5])]
                start_parameter = [minmag[0][0], minmag[0][1], 2.5, 3.0]
                try:
                    popt, pcov = curve_fit(utils.sk_2dprofile, currentlayer[:, :2], currentlayer[:, 5], start_parameter,
                                           maxfev=2000)
                    popt, pcov = curve_fit(utils.sk_2dprofile, currentlayer[:, :2], currentlayer[:, 5], popt)
                    rads.append(utils.sk_radius(popt[2], popt[3]))
                except RuntimeError:
                    print(f'layer {n} does not contain skyrmion.')
                    rads.append(None)
        return rads

    @property
    def topologies(self) -> List[ITopology]:
        r"""
        Returns a list of topology objects. One for each layer
        """
        if self.source != LATT_TYPE_SPIN:
            print('WARNING: You calc. the topology of a non-spin-lattice. Might not make sense at all.')
        self._topologies = []
        for idx_layer in range(self.nlayer):
            selected_layer = self.getlayer_by_idx(idx=idx_layer)
            points = selected_layer[:, :3]
            spins = selected_layer[:, 3:6]
            self._topologies.append(CTopologyHexCubic(points=points, spins=spins, N1=self.N1, N2=self.N2))
        return self._topologies



r"""
Construction class for translational modes
"""
from typing import Union, Tuple

import numpy as np

from spinterface.inputs.lattice.modeconstruction.imodeconstruction import IModeConstruction


class CConstructTranslational(IModeConstruction):
    r"""

    """

    def __init__(self, spins: np.ndarray, points: np.ndarray, lattvec: np.ndarray) -> None:
        r"""
        """
        super().__init__(spins, points, lattvec)
        self.shiftedspins = self._create_translated_spins()
        self.evec = self._create_evec()

    def _create_evec(self) -> np.ndarray:
        r"""
        Create translation vector based on the spin configuration and the shifted configuration
        """
        evec = []
        for index, spin in enumerate(self.spins):
            angle = np.arccos(np.dot(spin / np.linalg.norm(spin),
                                     self.shiftedspins[index] / np.linalg.norm(self.shiftedspins[index])))
            spin_diff = self.shiftedspins[index] - spin
            displacement = spin_diff - np.dot(spin_diff, spin) * spin
            displacement = displacement / np.linalg.norm(displacement)
            evec.append(displacement * angle)
        return np.asarray(evec)

    def _create_translated_spins(self) -> np.ndarray:
        r"""
        Saves for each spin the neighbor spin in direction of the next lattice vector
        """
        translated_spins = []
        # under the assumption of equal lattice size in both lattice vectors
        latt_size = int(np.sqrt(len(self.points)))
        for index, point in enumerate(self.points):
            searchpoint = point + self.lattvec
            # find index of neighbor
            index_searchpoint = np.argwhere((np.linalg.norm(self.points - searchpoint, axis=1)) <= 1.0e-4)
            # periodic boundaries
            # find all points on a line through the current point in the direction of lattvec
            if not index_searchpoint:
                searchpoint = point - (latt_size - 1) * self.lattvec
                index_searchpoint = np.argwhere((np.linalg.norm(self.points - searchpoint, axis=1)) <= 1.0e-4)
            translated_spins.append(self.spins[index_searchpoint[0][0]])
        return np.asarray(translated_spins)

    def __call__(self) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        return self.evec



r"""
Abstract class for functions which meet the requirements of the visualizer of the energy landscape. Inheriting from
this class for your own defined functions ensures this fulfillment.
"""
from abc import ABC, abstractmethod
import numpy as np


class I2DFunction(ABC):
    r"""
    Abstract base class for functions for visualizing energy landscapes on 2D spaces.
    """

    @abstractmethod
    def __call__(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        r"""
        Abstract method which prescribes the functions to implement for the use in the visualization interface.
        """


    #===============utility function which can be used to define a combination in the call-method when inheriting=======


    @staticmethod
    def gauss(x: np.ndarray, y: np.ndarray, x0: float, y0: float, a: float, sigmax: float, sigmay: float) -> np.ndarray:
        r"""
        Gaussian function
        Args:
            x: x-array
            y: y-array
            x0: x offset zero point
            y0: y offset zero point
            a: amplitude
            sigmax: x width
            sigmay: y width

        Returns:
            a * e^{-(x - x0)^2 / sigmax} * e^{-(y - y0)^2 / sigmay}

        """
        return a * np.exp(-(((x - x0) ** 2 / sigmax) + ((y - y0) ** 2 / sigmay ** 2)))

    @staticmethod
    def spherical_sin(x: np.ndarray, y: np.ndarray, omega: float, a: float):
        r"""
        Spherical sine
        Args:
            x: x-array
            y: y-array
            omega: frequency
            a: amplitude

        Returns:
            a * sin(w * sqrt(w * (x^2 + y^2))

        """
        return a * np.sin(omega * np.sqrt(omega * (x ** 2 + y ** 2)))



r"""
Input reader for input files in the style of the spinaker code.
"""
from pathlib import Path
from typing import Union, List, Any, Dict


class CInputReader:
    r"""
    Class for reading input files with the style of spinaker code. The keywords are separated from the values by a
    white space. Several values may follow a key. The default comment character is the #.
    """

    def __init__(self, fpath: Union[Path, str], commentchar: str = '#') -> None:
        r"""
        Args:
            fpath(Path,str): path to input file
        """
        self._fpath = fpath
        self._commentchar = commentchar
        self._fcontent = self._loadfilecontent()

    def _loadfilecontent(self) -> Dict[str, List]:
        r"""
        Loads the content of the file ignoring the lines with comment char.

        Returns:
            A dictionary containing the key value pairs of the input file. As there could be more than one value assigned
            to each key the values are represented by a list.
        """
        fcontent = {}
        with open(self._fpath) as f:
            for line in f:
                if line == '\n':
                    continue
                line_stripped = line.strip()
                if line_stripped.startswith(self._commentchar):
                    continue
                # remove trailing comments from line
                line_key_value = line_stripped.split(self._commentchar, 1)[0]
                line_splitted = line_key_value.split()
                # assume that the key is the first entry in each line
                currentkey = str(line_splitted[0])
                currentvalues = [str(val) for val in line_splitted[1:]]
                # convert bool values from fortran style into python style
                fcontent[currentkey] = currentvalues
        return fcontent

    @property
    def content(self) -> Dict[str, List]:
        r"""
        Returns:
            content of the input file
        """
        return self._fcontent



# -*- coding: utf-8 -*-

# @todo:ADAPT THIS TO THE NEW STRUCTURE
r"""
Naive implementation of the saddle point search for 2d (x,y).
"""
import pandas as pd
import numpy as np
from pathlib import Path
from typing import Callable


class CSPS_2D_naive:
    r"""
    Class for naive implementation of SPS
    """

    def __init__(self, f: Callable, initial: np.ndarray, iterations: int = 1000, ds: float = 0.001,
                 h: float = 0.00001) -> None:
        r"""
        Initialize SPS

        Args:
            f(Callable): Callable for energy landscape

            initial(np.ndarray): start point for the calculation

            iterations(int): number of eigenvector displacements

            ds(float): step width along eigenvector

            h(float): finite element

        """
        if len(initial) != 2:
            raise ValueError('Start point has to be 2 dimensional')
        self._initial = initial
        self._f = f
        self._iterations = iterations
        self._ds = ds
        self._h = h

    def _gradient(self, point: np.ndarray, h: float) -> np.ndarray:
        r"""
        Calculate the energy gradient

        Args:
            point(np.ndarray): Point for which the gradient is evaluated.
            h(float): finite element

        Return:
             1d numpy array with length 2: [dfdx,dfdy]
        """
        dfdx = (self._f(point + h * np.array([1, 0])) - self._f(point)) / h
        dfdy = (self._f(point + h * np.array([0, 1])) - self._f(point)) / h
        return np.array([dfdx, dfdy])

    def _hessian(self, point: np.ndarray, h: float) -> np.ndarray:
        r"""
        Calculate the hessian for a given point with finite elements

        Args:
            point(np.ndarray): Point for which the hessian is evaluated.
            h(float): finite element

        Returns:
            numpy array of hessian: [[dfdxdx, dfdxdy],[dfdydx, dfdydy]]
        """
        hess = np.zeros(shape=(2, 2))
        # diagonal terms
        hess[0, 0] = (self._f(point + 2 * h * np.array([1, 0])) - 2 * self._f(point + h * np.array([1, 0])) + self._f(
            point)) / h ** 2
        hess[1, 1] = (self._f(point + 2 * h * np.array([0, 1])) - 2 * self._f(point + h * np.array([0, 1])) + self._f(
            point)) / h ** 2
        hess[0, 1] = (self._f(point + h * np.array([1, 1])) - self._f(point + h * np.array([1, 0])) - self._f(
            point + h * np.array([0, 1])) + self._f(point)) / h ** 2
        hess[1, 0] = hess[0, 1]
        return hess

    def __call__(self, gralog: int, write_out: Path = Path.cwd() / 'sps.dat'):
        r"""
        Call the SPS

        Args:
            gralog(int): frequency of output writing
        """
        current_point = self._initial
        with open(str(write_out), 'w') as f:
            f.write(f'iter  pointx  pointy  eval1  eval2  evec1x  evec1y  evec2x  evec2y  forcex  forcey\n')
        for iter in range(self._iterations):
            hess = self._hessian(current_point, self._h)
            # calculate eigenvalues and normalized eigenvectors:
            eigenvals, eigenvecs = np.linalg.eigh(hess)
            # eigenvalues and eigenvectors are returned in ascending order. Chose minimum:
            eigenval = eigenvals[0]
            eigenvec = eigenvecs[:, 0]
            # calculate negative gradient
            neg_grad = -1 * self._gradient(current_point, self._h)
            force = neg_grad - 2 * (np.dot(neg_grad, eigenvec)) * eigenvec
            # do step (eventually scale with eigenvalue?)
            current_point = current_point + force * self._ds
            if np.mod(iter, gralog) == 0:
                with open(str(write_out), 'a') as f:
                    f.write(
                        f'{iter}  {current_point[0]}  {current_point[1]}  {eigenvals[0]}  {eigenvals[1]}  {eigenvecs[0, 0]}  {eigenvecs[1, 0]}  {eigenvecs[0, 1]}  {eigenvecs[1, 1]}  {force[0]}  {force[1]}\n')
                print(f"iteration GNEB: {iter},point: {current_point} , force: {force}")



# -*- coding: utf-8 -*-
r"""
This subpackage of the spinterface includes python implementations of algorithm which final versions may or may not be
implemented in the spin dynamic program. They can represent straightforward implementations without optimization written
in easy readable python code or test cases for low dimensionalities. These algorithms are also sometimes the blueprint
for a serious fortran implementation. Also they can be used to demonstrate the way the algorithms work. In this context
the sub-package visualizations.energylandscape is of interest since with that it's possible to visualize the behaviour
of these algorithms on a low dimensional energy surface. So far this package contains the following algorithms:

========================================================================================================================



"""


from spinterface.inputs.lattice.CLattice import CLattice
from pathlib import Path
from spinterface.visualizations.lattices.cvisualpyvista import CVisualPyVista

latt = CLattice(source='evec',path=Path.cwd() / 'vec_4_skyr.dat')

visual = CVisualPyVista(lattice=latt,cam=[(53.486631523554976, 28.986632387822333, 28.986613940184775),
 (24.49999913573265, 0.0, -1.84476375579834e-05),
 (0.0, 0.0, 1.0)])
visual.show()
visual(outpath=Path.cwd() / 'vec_4_skyr.png')


# -*- coding: utf-8 -*-
r"""
contains constants for visualization
"""
SPINDEFAULT_TIPLENGTH: float = 1.0
SPINDEFAULT_TIPRADIUS: float = 0.35
SPINDEFAULT_ARROWSCALE: float = 0.7
SPINDEFAULT_DRAWBACKGROUND: bool = False
SPINDEFAULT_SETTINGS={'tiplength':SPINDEFAULT_TIPLENGTH,'tipradius':SPINDEFAULT_TIPRADIUS,'arrowscale':SPINDEFAULT_ARROWSCALE, 'drawbackground':SPINDEFAULT_DRAWBACKGROUND}

EVECDEFAULT_TIPLENGTH: float = 0.4
EVECDEFAULT_TIPRADIUS: float = 0.25
EVECDEFAULT_ARROWSCALE: float = 5.0
EVECDEFAULT_DRAWBACKGROUND: bool = True
EVECDEFAULT_SETTINGS={'tiplength':EVECDEFAULT_TIPLENGTH,'tipradius':EVECDEFAULT_TIPRADIUS,'arrowscale':EVECDEFAULT_ARROWSCALE, 'drawbackground':EVECDEFAULT_DRAWBACKGROUND}

FORCEDEFAULT_TIPLENGTH: float = 0.4
FORCEDEFAULT_TIPRADIUS: float = 0.25
FORCEDEFAULT_ARROWSCALE: float = 10000.0
FORCEDEFAULT_DRAWBACKGROUND: bool = True
FORCEDEFAULT_SETTINGS={'tiplength':FORCEDEFAULT_TIPLENGTH,'tipradius':FORCEDEFAULT_TIPRADIUS,'arrowscale':FORCEDEFAULT_ARROWSCALE, 'drawbackground':FORCEDEFAULT_DRAWBACKGROUND}



# -*- coding: utf-8 -*-
r"""
Abstract interface class for visualizing energy landscapes of functions defined on a two dimensional configuration
space. These visualizations can be used to explain the methods performed by the Spin Dynamics Code on high dimensional
configuration space. E.G. the GNEB or ci-GNEB method, the saddle point search, the mode following or even minimization
procedures. In that context the package spineapple included in this package here can be used (It includes low dim.
implementations of the algorithms of the spindynamic code).
"""
from abc import ABC, abstractmethod
from typing import Callable, Union
from spinterface.visualizations.energylandscapes.i2dfunction import I2DFunction
import numpy as np


class IVisualizer(ABC):
    r"""
    Abstract base class for visualizing functions on two dimensional configuration spaces.
    """

    def __init__(self, x: np.ndarray, y: np.ndarray, function: I2DFunction) -> None:
        r"""
        Initializer for the Visualizer.

        Args:
            x(np.array): x array - can be created with linspace (together with y a meshgrid will be created.
            y(np.array): y array - can be created with linspace (together with x a meshgrid will be created.
            function(I2DFunction): function defined on a two dimensional configuration space.
        """
        self.func = function
        self.x = x
        self.y = y
        self.xgrid, self.ygrid = np.meshgrid(self.x, self.y)
        # evaluate the
        self.zgrid = function(self.xgrid, self.ygrid)



from abc import abstractmethod, ABC
from spinterface.inputs.lattice.CLattice import CLattice
from spinterface.visualizations.lattices.cvisualpyvista import CVisualPyVista
from typing import List
from pathlib import Path


class IVisChain(ABC):
    r"""
    Chain for visualization tasks
    """

    def __init__(self, ) -> None:
        r"""
        Initializes the visualization chain
        """


r"""
Visualizer for the esc-stage of the sps algorithm
"""
from spinterface.sps_results.stage_visualizer.ivisstage import IVisStage
from otterplot.model.cotterplotter import COtterPlotter
import matplotlib.pyplot as plt
from pathlib import Path
from typing import Union, Tuple, List
from spinterface.inputs.lattice.CLattice import CLattice
from spinterface.visualizations.lattices.cvisualpyvista import CVisualPyVista
import pandas as pd


class CVisConv(IVisStage):
    r"""
    Visualization class for convergence stage
    """

    def __call__(self) -> None:
        r"""
        Excecutes the visualization for the converge stage
        """
        if self.strategy in [1, 2]:
            self._visualize_converged_configuration()
            self._visualize_eigenvector_start_end()

    def __init__(self, strategy: int, info_file: Union[Path, str],
                 camera_spin_files: Union[List[Tuple[float, float, float]], None] = None,
                 spinvisualization_technique: str = 'oop', arrowscale_evec: float = 30) -> None:
        r"""
        Initializes the visualizer for the converge-stage of SPS
        Args:
            strategy(int): converge strategy
            info_file(Path,str): the file which holds information about the locations
            of the corresponding files. It is always in the same directory as the sps.in therefore all paths in the info
            file are relative to this.
        """
        super().__init__(strategy, camera_spin_files)
        self.parent_directory = Path(info_file).parent
        self.arrowscale_evec = arrowscale_evec
        self.keys, self.esckeys, self.sp_or_convex, self.folders, self.spinfiles = self._read_info_file(
            info_file=info_file)
        self.spinvisualization_technique = spinvisualization_technique

    @staticmethod
    def _read_info_file(info_file: Union[Path, str]) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]:
        r"""
        Reads the columns of the converge-info file.
        """
        df = pd.read_csv(info_file, sep=r'\s+')
        return df['DISPKEY'], df['ESCKEY'], df['sp_or_convex'], df['folder'], df['spin_file']

    @property
    def stage(self) -> str:
        r"""
        Returns:
            the stage
        """
        return 'Converge'

    def _stagestrategytext(self) -> str:
        r"""
        Returns:
            the converge strategy in words
        """
        if self.strategy == 1:
            return 'Use minimummodefollowing to converge on saddlepoint.'
        elif self.strategy == 2:
            return 'Use mfef (mode following with inversion) to converge on saddlepoint.'
        else:
            return 'Describing text for this strategy not yet coded.'

    def get_info_files(self) -> List[Path]:
        r"""
        returns:
            A list of the paths to the info files
        """
        info_files = []
        for spinfile in self.spinfiles:
            p = self.parent_directory / spinfile
            if self.strategy == 1:
                p = p.parent / 'info_mmf.dat'
            elif self.strategy == 2:
                p = p.parent / 'info_mfef.dat'
            info_files.append(p)
        return info_files

    def _visualize_converged_configuration(self):
        r"""
        Visualizes the spin configurations for all converged or re-entered configurations
        """
        for spinfile in self.spinfiles:
            print(f'Visualizig: {spinfile}...')
            l_latt = CLattice(source='STM', path=self.parent_directory / spinfile)
            if self.spinvisualization_technique == 'oop':
                l_visu = CVisualPyVista(lattice=l_latt, cam=self.spincam)
            elif self.spinvisualization_technique == 'ipoop':
                l_visu = CVisualPyVista(lattice=l_latt, cam=self.spincam, heatmap=True, heatmap_saturation=1.0)
            else:
                raise ValueError('Selected spin visualization technique is not coded yet!')
            pngfile = spinfile.replace('.dat', '.png')
            pngfile_splitted = pngfile.split('/')
            pngpath = self.parent_directory
            for pathpart in pngfile_splitted:
                pngpath = pngpath / pathpart
            l_visu(outpath=pngpath)
            l_visu.plotter.close()
            print('...done')

    def _visualize_eigenvector_start_end(self):
        r"""
        Visualizes the spin configurations for all converged configurations
        """
        for spinfile in self.spinfiles:
            evecfile_end = spinfile.replace('spin', 'evec')
            print(f'Visualizig: {evecfile_end}...')
            l_latt_end = CLattice(source='evec', path=self.parent_directory / evecfile_end)
            l_visu_end = CVisualPyVista(lattice=l_latt_end, cam=self.spincam, arrowscale=self.arrowscale_evec)
            pngfile_end = evecfile_end.replace('.dat', '.png')
            pngfile_splitted_end = pngfile_end.split('/')
            pngpath_end = self.parent_directory
            for pathpart in pngfile_splitted_end:
                pngpath_end = pngpath_end / pathpart
            l_visu_end(outpath=pngpath_end)
            l_visu_end.plotter.close()
            print('...done')
            evecfile_start = evecfile_end.replace('end', '0')
            print(f'Visualizig: {evecfile_start}...')
            l_latt_start = CLattice(source='evec', path=self.parent_directory / evecfile_start)
            l_visu_start = CVisualPyVista(lattice=l_latt_start, cam=self.spincam, arrowscale=self.arrowscale_evec)
            pngfile_start = evecfile_start.replace('.dat', '.png')
            pngfile_splitted_start = pngfile_start.split('/')
            pngpath_start = self.parent_directory
            for pathpart in pngfile_splitted_start:
                pngpath_start = pngpath_start / pathpart
            l_visu_start(outpath=pngpath_start)
            l_visu_start.plotter.close()
            print('...done')



# -*- coding: utf-8 -*-
r"""
This module contains utilities for visualizations of spin configurations.
"""
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.colors as colors
import numpy as np


def truncate_colormap(cmap, minval: float = 0.0, maxval: float = 1.0, n: int = 100):
    r"""
    Returns:
        A new colormap which is a segment of a known color map in matplotlib
    """
    return colors.LinearSegmentedColormap.from_list(
        'trunc({c},{a:.2f},{b:.2f})'.format(c=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))


def get_colormap(which: str = 'hsv_spind') -> mpl.colors.LinearSegmentedColormap:
    r"""
    Returns:
        the desired colormap based on a key string.
    """
    if which == 'hsv_spind':
        mpl.cm.register_cmap('hsv_new', cmap=truncate_colormap(mpl.cm.get_cmap('hsv'), minval=0.6, maxval=0.0))
        return mpl.cm.get_cmap('hsv_new')
    elif which == 'paraview_standard':
        return mpl.cm.get_cmap('viridis')
    elif which == 'coolwarm':
        return mpl.cm.get_cmap('coolwarm')


def HSVtoRGB(vec: np.ndarray, saturation: float = 0.2) -> np.ndarray:
    r"""
    Calculates the bivariate color code varying hue and value for vectors. The oop-component will be visualized by
    the value and the ip-angle by hue. The saturation can be provided as input
    """
    vecnorm = np.linalg.norm(vec)
    H = np.arctan2(vec[1] / vecnorm, vec[0] / vecnorm)
    value = 1 - abs(vec[2] / vecnorm)
    chroma = saturation * value
    # H angle is in interval -pi;pi shift it to 0,2pi
    # also transfer to 0;360
    H = (H + np.pi) * (180 / np.pi)
    Hprime = H / 60
    X = chroma * (1 - abs(Hprime % 2 - 1))
    if 0 <= Hprime < 1:
        r1 = chroma
        g1 = X
        b1 = 0
    elif 1 <= Hprime < 2:
        r1 = X
        g1 = chroma
        b1 = 0
    elif 2 <= Hprime < 3:
        r1 = 0
        g1 = chroma
        b1 = X
    elif 3 <= Hprime < 4:
        r1 = 0
        g1 = X
        b1 = chroma
    elif 4 <= Hprime < 5:
        r1 = X
        g1 = 0
        b1 = chroma
    elif 5 <= Hprime <= 6:
        r1 = chroma
        g1 = 0
        b1 = X
    else:
        raise ValueError
    m = value - chroma
    r = r1 + m
    g = g1 + m
    b = b1 + m
    return np.array([r, g, b])



r"""
Visualization class for initial stage of algorithm
"""
from spinterface.sps_results.stage_visualizer.ivisstage import IVisStage
from pathlib import Path
from typing import Union, Tuple, List
from spinterface.inputs.lattice.CLattice import CLattice
from spinterface.visualizations.lattices.cvisualpyvista import CVisualPyVista


class CVisInitial(IVisStage):
    r"""
    Visualize initial state
    """

    def __init__(self, strategy: int, info_file: Union[Path, str],
                 camera_spin_files: Union[List[Tuple[float, float, float]], None] = None,
                 spinvisualization_technique: str = 'oop', i_initial_evecs: bool = True,
                 n_initial_evecs: int = 10, arrowscale_evec:float = 30) -> None:
        r"""
        Initializes visiualization stage

        Args:
            strategy(int): strategy
            info_file(Path,str): the file which holds information about the initial state
        """
        super().__init__(strategy, camera_spin_files)
        self.parent_directory = Path(info_file).parent
        self.info_file = info_file
        self.i_initial_evecs = i_initial_evecs
        self.n_initial_evecs = n_initial_evecs
        self.arrowscale_evec = arrowscale_evec
        self.spinvisualization_technique = spinvisualization_technique

    @property
    def stage(self) -> str:
        r"""
        Returns:
            the stage
        """
        return 'Initial'

    def _stagestrategytext(self) -> str:
        r"""
        Returns:
            the displace strategy in words
        """
        if self.strategy == 1:
            return 'Visualize eigenvectors of initial state'
        else:
            return 'Describing text for this strategy not yet coded.'

    def __call__(self) -> None:
        r"""
        Excecutes the visualization for the initial stage
        """
        if self.strategy in [1]:
            if not self.i_initial_evecs:
                raise ValueError('Strategy 1 cant be chosen if i_initial_eigenvec in input sps file is False.')
            self._visualize_eigenvectors()

    def _visualize_eigenvectors(self):
        r"""
        Visualizes the spin configurations for all displaced configurations
        """
        for n in range(1,self.n_initial_evecs+1):
            p = self.parent_directory / 'initial_eigenvectors' / f'evec_initial_{n}.dat'
            p_png = self.parent_directory / 'initial_eigenvectors' / f'evec_initial_{n}.png'
            l_latt = CLattice(source='evec', path=p)
            l_visu = CVisualPyVista(lattice=l_latt, cam=self.spincam, arrowscale=self.arrowscale_evec)
            l_visu(outpath=p_png)
            l_visu.plotter.close()



r"""
Based on a spin lattice construct a normalized eigenvector which is a mode for this lattice (most likely this will be
realized as a translation or rotation).
"""
from abc import ABC, abstractmethod
from typing import Union, Tuple
import numpy as np


class IModeConstruction(ABC):
    r"""
    Abstract base class for construction of modes for selected lattices.
    """
    def __init__(self, spins: np.ndarray, points: np.ndarray, lattvec: np.ndarray) -> None:
        r"""
        Args:
            spins: spins of the lattice
            points: lattice coordinates of the spins
        """
        self.spins = spins
        self.points = points
        self.lattvec = lattvec

    @abstractmethod
    def __call__(self) -> Union[Tuple[np.ndarray,np.ndarray],np.ndarray]:
        r"""
        Returns:
            eigenvector or tuple of eigenvectors
        """



r"""
Main class for defining a result-visualization for a saddle point search implementation
"""
from pathlib import Path
from spinterface.inputs.lattice.CLattice import CLattice
from spinterface.visualizations.lattices.cvisualpyvista import CVisualPyVista
from typing import Union, Dict, List, Tuple, Any
from spinterface.sps_results.inputreader.CInputReader import CInputReader
import spinterface.sps_results.constants_CSPS as const
from spinterface.sps_results.stage_visualizer.cvisdisp import CVisDisp
from spinterface.sps_results.stage_visualizer.cvisesc import CVisEsc
from spinterface.sps_results.stage_visualizer.cvisconv import CVisConv
from spinterface.sps_results.stage_visualizer.cvistree import CVisTree
from spinterface.sps_results.stage_visualizer.cvisinitial import CVisInitial
from spinterface.sps_results.stage_visualizer.ivisstage import IVisStage


class CSPS:
    r"""
    Class which summarizes a bunch of visualizations for sps calculations.
    """

    def __init__(self, inputfile: Union[Path, str] = Path.cwd() / 'sps.in', stage: str = 'DISP',
                 info_initial: Union[Path, str] = Path.cwd() / 'info_sps_eigenvalues.dat',
                 info_disp: Union[Path, str] = Path.cwd() / 'info_sps_disp.dat',
                 info_esc: Union[Path, str] = Path.cwd() / 'info_sps_esc.dat',
                 info_conv: Union[Path, str] = Path.cwd() / 'info_sps_conv.dat',
                 cameras_spin_files: Union[List[Tuple[float, float, float]], None] = None,
                 spinvisualization_technique: str = 'oop',
                 arrowscale_evec: float = 30) -> None:
        r"""
        Initializes SPS-Visualization
        """
        self.arrowscale_evec = arrowscale_evec
        self.inputfile = inputfile
        self.info_initial = info_initial
        self.info_disp = info_disp
        self.info_esc = info_esc
        self.info_conv = info_conv
        self.spinvisualization_technique = spinvisualization_technique
        l_inp_reader = CInputReader(fpath=inputfile)
        self._cameras_spin_files = cameras_spin_files
        self.sps_settings = self._check_settings(l_inp_reader.content)
        self._iinitial, self._idisp, self._iesc, self._iconv, self._iesca, self._itree = self._parse_stage(stage)
        self.stages = self._create_visualizer()

    def __call__(self) -> None:
        r"""
        Executes the visualizations
        """
        for (stagename, stage) in self.stages.items():
            print(f'Visualizing stage: {stagename}')
            stage()

    def stage_visualizer(self, name: str) -> IVisStage:
        r"""
        Gets the stage visualizer to access the methods of these classes directly
        """
        try:
            stage = self.stages[name]
        except KeyError:
            raise ValueError(f'Stage {name} not initialized. Current stages available are: {self.stages.keys()}.')
        return stage

    def show_initial_structure(self, topview: bool = False):
        r"""
        shows initial structure
        """
        # Visualize the initial configuration
        l_latt = CLattice(source='STM', path=self.info_disp.parent / 'SpinSTMi.dat')
        if self.spinvisualization_technique == 'oop':
            l_visu = CVisualPyVista(lattice=l_latt, cam=self._cameras_spin_files)
        elif self.spinvisualization_technique == 'ipoop':
            l_visu = CVisualPyVista(lattice=l_latt, cam=self._cameras_spin_files, heatmap=True, heatmap_saturation=1.0)
        else:
            raise ValueError('Selected spin visualization technique is not coded yet!')
        pngfile = self.info_disp.parent / 'SpinSTMi.png'
        if topview:
            l_visu.plotter.view_xy()
        l_visu.show()
        l_visu(outpath=pngfile)
        l_visu.plotter.close()

    def _create_visualizer(self) -> Dict[str, IVisStage]:
        r"""
        Initializes the visualizers for the tasks
        """
        l_stages = {}
        if self._iinitial:
            l_stages['INITIAL'] = CVisInitial(strategy=1, info_file=self.info_initial,
                                              camera_spin_files=self._cameras_spin_files,
                                              spinvisualization_technique=self.spinvisualization_technique,
                                              i_initial_evecs=self.sps_settings['i_write_initial_eigenvecs'][0],
                                              n_initial_evecs=self.sps_settings['i_write_initial_eigenvecs'][1],
                                              arrowscale_evec=self.arrowscale_evec)
        if self._idisp:
            l_stages['DISP'] = CVisDisp(strategy=self.sps_settings['algo_displace'], info_file=self.info_disp,
                                        camera_spin_files=self._cameras_spin_files,
                                        spinvisualization_technique=self.spinvisualization_technique)
        if self._iesc:
            l_stages['ESC'] = CVisEsc(strategy=self.sps_settings['algo_escape'], info_file=self.info_esc,
                                      camera_spin_files=self._cameras_spin_files,
                                      spinvisualization_technique=self.spinvisualization_technique,
                                      arrowscale_evec=self.arrowscale_evec)
        if self._iconv:
            l_stages['CONV'] = CVisConv(strategy=self.sps_settings['algo_converge'], info_file=self.info_conv,
                                        camera_spin_files=self._cameras_spin_files,
                                        spinvisualization_technique=self.spinvisualization_technique,
                                        arrowscale_evec=self.arrowscale_evec)
        if self._itree:
            l_stages['TREE'] = CVisTree(strategy=2, info_file_disp=self.info_disp, info_file_esc=self.info_esc,
                                        info_file_conv=self.info_conv,
                                        stage_strategies={'DISP': self.sps_settings['algo_displace'],
                                                          'ESC': self.sps_settings['algo_escape'],
                                                          'CONV': self.sps_settings['algo_converge'],
                                                          'ESCA': self.sps_settings['algo_escape']})
        return l_stages

    @staticmethod
    def _parse_stage(stage: str) -> Tuple[bool, bool, bool, bool, bool, bool]:
        r"""
        Parses the parse information decider string.
        """
        if stage not in ['ALL', 'INITIAL', 'DISP', 'ESC', 'CONV', 'ESCA', 'TREE']:
            raise ValueError(f'Stage: {stage} not in allowed options (ALL, DISP, ESC, CONV, ESCA, TREE).')
        if stage == 'ALL':
            return True, True, True, True, True, True
        else:
            return stage == 'INITIAL', stage == 'DISP', stage == 'ESC', stage == 'CONV', stage == 'ESCA', stage == 'TREE'

    @staticmethod
    def _check_settings(settings: Dict[str, List]) -> Dict[str, Any]:
        r"""
        Checks if all the necessary settings are available and convert the data types.
        """
        checked_settings = {}
        for (key, valuetype) in const.KEYS.items():
            try:
                l_values = settings[key]
            except KeyError:
                raise ValueError(f"The following necessary key is not provided in input file: {key}.")
            if len(l_values) == 1:
                l_value = l_values[0]
                if valuetype == float:
                    l_value = l_value.replace('d', 'e')
                if valuetype == bool:
                    if l_value in ['.T.', '.True.']:
                        checked_settings[key] = True
                        continue
                    if l_value in ['.F.', '.False.']:
                        checked_settings[key] = False
                        continue
                checked_settings[key] = valuetype(l_value)
            elif len(l_values) == 2:
                # in this case valuetype is a list
                vals = []
                for index, vt in enumerate(valuetype):
                    act_val = l_values[index]
                    if vt == float:
                        act_val = act_val.replace('d', 'e')
                    if vt == bool:
                        if act_val in ['.T.', '.True.']:
                            act_val = True
                        if act_val in ['.F.', '.False.']:
                            act_val = False
                    vals.append(vt(act_val))
                checked_settings[key] = vals
        return checked_settings




from spinterface.inputs.CWriteInputs import CWriteInput
from spinterface.inputs.lattice.CLattice import CLattice
from spinterface.visualizations.lattices.cvisualpyvista import CVisualPyVista
from pathlib import Path
import numpy as np

lattice = CWriteInput(name='lattice.in', Nsize=f'20 20 1', alat='2.0 1.0 10.0',
                      lines=['lattice', '1.0 0.0 0.0', '0.0 1.0 0.0', '0.0 0.0 1.0',
                             'motif 2 atoms', f'0.0 0.0 0.0 {2.5}', '0.5 0.5 0.5 2.0'])
lattice(where=Path.cwd())

#latt = CLattice(source='STM',path=Path.cwd()/'SpinSTMi.dat')
latt = CLattice(magdir=np.array([1.0,0.0,0.0]))

latt.add_skyrmiontube(c=7.0,w=7.5)

latt.write()
print(latt.skradius)

visualizer = CVisualPyVista(lattice=latt)
#, cam=[(83.05595689117641, 66.51948874605938, 7.739481461469535),
# (19.499999448657036, 9.749999903142452, 0.24698486924171448),
# (-0.05425875800911427, -0.07071068929493245, 0.9960200728894695)])
visualizer.show()
#visualizer()


# -*- coding: utf-8 -*-
r"""
Module contains implementation of pyvista visualizations for spin lattices.
"""
from pathlib import Path

import matplotlib.pyplot as plt
import vtkmodules.util.vtkConstants
import matplotlib as mpl
from spinterface.visualizations.lattices.utilities import get_colormap
import pyvista as pv
import vtk
import numpy as np
from spinterface.visualizations.lattices.ivisualizer import IVisualizer
from spinterface.inputs.lattice.ILattice import ILattice
from typing import List, Tuple, Union, Dict
from spinterface.visualizations.const import SPINDEFAULT_SETTINGS, EVECDEFAULT_SETTINGS
from spinterface.inputs.lattice.const import LATT_TYPE_EVEC, LATT_TYPE_SPIN


class CVisualPyVistaNewCmap(IVisualizer):
    r"""
    Class for visualizing spin lattices with py vista library
    """

    def __init__(self, lattice: ILattice, tiplength: Union[float, None] = None, tipradius: Union[float, None] = None,
                 arrowscale: Union[float, None] = None, draw_background: Union[bool, None] = None,
                 cam: Union[List[Tuple[float, float, float]], None] = None,
                 cmap: str = 'hsv_spind', total_mag=False, topology=False) -> None:
        r"""
        Initializes the visualization

        Args:
            tiplength(float): geometry of arrow: tiplength
            tipradius(float): geometry of arrow: tipradius
            arrowscale(float): geometry of arrow: arrowscale
            draw_background(bool): shall i draw the background of the lattice
            camera: camera position
            cmap: string for the choice of the colormap. Defined in utilities module
            total_mag(bool): if the sum over the lattice vectors shall be visualized in the center of the lattice
            topology(bool): if topology information shall be visualized
        """
        super().__init__(lattice)
        self.text_objects = []
        self.total_mag = total_mag
        self.topology = topology
        self.tiplength, self.tipradius, self.arrowscale, self.drawbackground = self._load_settings(tiplength, tipradius,
                                                                                                   arrowscale,
                                                                                                   draw_background)
        self.cam = cam
        self.cmap = get_colormap(cmap)
        self._make_plotter()

    def _load_settings(self, tl: Union[float, None], tr: Union[float, None],
                       asc: Union[float, None], dbg: Union[bool, None]) -> Tuple[float, float, float, bool]:
        r"""
        Returns:
            loads the tiplength, tipradius and arrowscale depending on the inputs and the lattice type
        """
        # Decide on loading settings
        if self.lattice.source == LATT_TYPE_SPIN:
            print(f'loading defaults for type: {LATT_TYPE_SPIN}')
            tiplength = SPINDEFAULT_SETTINGS['tiplength']
            tipradius = SPINDEFAULT_SETTINGS['tipradius']
            arrowscale = SPINDEFAULT_SETTINGS['arrowscale']
            drawbackground = SPINDEFAULT_SETTINGS['drawbackground']
        elif self.lattice.source == LATT_TYPE_EVEC:
            print(f'loading defaults for type: {LATT_TYPE_SPIN}')
            tiplength = EVECDEFAULT_SETTINGS['tiplength']
            tipradius = EVECDEFAULT_SETTINGS['tipradius']
            arrowscale = EVECDEFAULT_SETTINGS['arrowscale']
            drawbackground = EVECDEFAULT_SETTINGS['drawbackground']
        else:
            raise ValueError('Not a valid lattice source!')
        if tl is not None:
            print('Overwriting tiplength setting with user input')
            tiplength = tl
        if tr is not None:
            print('Overwriting tiplradius setting with user input')
            tipradius = tr
        if asc is not None:
            print('Overwriting arrowscale setting with user input')
            arrowscale = asc
        if dbg is not None:
            print('Overwriting drawbackground setting with user input')
            drawbackground = dbg
        return tiplength, tipradius, arrowscale, drawbackground

    def _make_plotter(self, offscreen: bool = False):
        r"""
        Creates the plotter. The plotter will be recreated when saving the image
        """
        self.plotter = pv.Plotter(off_screen=offscreen, lighting='three lights')
        self._configureplotter()
        self.plotter.camera_position = self.cam
        plotpoints = self.lattice.points[:, :]
        plotspins = self.lattice.spins[:, :]
        plot_colors = self.get_colors(plotspins=plotspins)
        # create color dictionary
        # cmap = self.create_cmap(plot_colors)
        self.PD = pv.PolyData(plotpoints)
        self.PD["colors"]=plot_colors
        self.surf = self.PD.delaunay_2d()
        self.plotter.add_mesh(self.PD,scalars="colors",rgb=True)
        if self.total_mag:
            self._draw_total_magnetization()
        if self.topology:
            self._draw_topology()
        if self.drawbackground:
            self._draw_background()

    def create_cmap(self, plot_colors) -> mpl.colors.ListedColormap:
        r"""

        """
        N = len(plot_colors)  # number of points
        print(N)
        vals = np.ones((N, 4))
        vals[:, 0] = plot_colors[:, 0]
        vals[:, 1] = plot_colors[:, 1]
        vals[:, 2] = plot_colors[:, 2]
        cmap = mpl.colors.ListedColormap(vals)
        return cmap

    def get_colors(self, plotspins: np.ndarray) -> np.ndarray:
        r"""

        """
        colors = []
        for spin in plotspins:
            colors.append(self.HSVtoRGB(spin))
        return np.array(colors)

    @staticmethod
    def HSVtoRGB(vec: np.ndarray, saturation: float = 0.2) -> np.ndarray:
        r"""

        """
        vecnorm = np.linalg.norm(vec)
        H = np.arctan2(vec[1] / vecnorm, vec[0] / vecnorm)
        value = 1 - abs(vec[2] / vecnorm)
        chroma = saturation * value
        # H angle is in interval -pi;pi shift it to 0,2pi
        # also transfer to 0;360
        H = (H + np.pi) * (180 / np.pi)
        Hprime = H / 60
        X = chroma * (1 - abs(Hprime % 2 - 1))
        if 0 <= Hprime < 1:
            r1 = chroma
            g1 = X
            b1 = 0
        elif 1 <= Hprime < 2:
            r1 = X
            g1 = chroma
            b1 = 0
        elif 2 <= Hprime < 3:
            r1 = 0
            g1 = chroma
            b1 = X
        elif 3 <= Hprime < 4:
            r1 = 0
            g1 = X
            b1 = chroma
        elif 4 <= Hprime < 5:
            r1 = X
            g1 = 0
            b1 = chroma
        elif 5 <= Hprime <= 6:
            r1 = chroma
            g1 = 0
            b1 = X
        else:
            raise ValueError
        m = value - chroma
        r = r1 + m
        g = g1 + m
        b = b1 + m
        return np.array([r, g, b])

    def _draw_topology(self) -> None:
        r"""
        Visualizes the topology information of a lattice.
        """
        if self.lattice.source != LATT_TYPE_SPIN:
            print('WARNING: topology for non spin lattices might not make sense.')
        topos = self.lattice.topologies
        # draw the topologic density for each layer
        for layeridx in range(self.lattice.nlayer):
            topo = topos[layeridx]
            cell_types = np.array([vtk.VTK_TRIANGLE for count in range(len(topo.polygons1[::3, 1]))], np.int8)
            cells_list = []
            for idx in range(len(topo.polygons1[::3, 1])):
                cells_list.extend([3, idx * 3, idx * 3 + 1, idx * 3 + 2])
            cells = np.asarray(cells_list)
            grid1 = pv.UnstructuredGrid(cells, cell_types, topo.polygons1[:, :])
            grid1['dens'] = topo.areas1
            self.plotter.add_mesh(grid1, cmap='seismic')
            cell_types = np.array([vtk.VTK_TRIANGLE for count in range(len(topo.polygons2[::3, 1]))], np.int8)
            cells_list = []
            for idx in range(len(topo.polygons2[::3, 1])):
                cells_list.extend([3, idx * 3, idx * 3 + 1, idx * 3 + 2])
            cells = np.asarray(cells_list)
            grid2 = pv.UnstructuredGrid(cells, cell_types, topo.polygons2[:, :])
            grid2['dens'] = topo.areas2
            self.plotter.add_mesh(grid2, cmap='seismic')
            # visualize the center of mass with respect to topologic density.
            center = pv.PolyData(topo.topological_center)
            center = pv.Sphere(radius=1.0, center=topo.topological_center)
            # self.plotter.add_mesh(center, color='k')

    def _draw_total_magnetization(self) -> None:
        r"""
            the total magnetization (the sum of all spins) in the center of the lattice
        """
        length_tot_mag = np.linalg.norm(self.lattice.total_magnetization)

        if self.lattice.source == 'evec':
            if length_tot_mag < 50:
                print('WARNING: the norm of the total magnetization is below 50. The visualization of it might not be '
                      'visible!')
            arrow = pv.Arrow(start=self.lattice.midpoint, direction=self.lattice.total_magnetization, tip_length=0.25,
                             tip_radius=0.1, tip_resolution=20,
                             shaft_radius=0.05, shaft_resolution=20, scale=length_tot_mag / 10)
        else:
            if length_tot_mag < 500:
                print('WARNING: the norm of the total magnetization is below 500. The visualization of it might not be '
                      'visible!')
            arrow = pv.Arrow(start=self.lattice.midpoint, direction=self.lattice.total_magnetization, tip_length=0.25,
                             tip_radius=0.1, tip_resolution=20,
                             shaft_radius=0.05, shaft_resolution=20, scale=length_tot_mag / 100)
        self.plotter.add_mesh(arrow, color='k')

    def _draw_background(self) -> None:
        r"""
        Draws the background of the lattice
        """
        for layer in range(self.lattice.nlayer):
            magstruct = self.lattice.getlayer_by_idx(layer)
            points = magstruct[:, :3]
            points_poly = pv.PolyData(points)
            surface = points_poly.delaunay_2d()
            self.plotter.add_mesh(surface, show_edges=True, opacity=0.5)


    def _configureplotter(self) -> None:
        r"""
        Configures the plotter object
        """
        pv.set_plot_theme("ParaView")
        pv.rcParams['transparent_background'] = True
        self.plotter.set_background('black')

        def cam() -> None:
            print('Camera postion: ', self.plotter.camera_position)

        self.plotter.add_key_event('c', cam)

    def show(self) -> None:
        r"""
        Shows the plotter
        """
        print('Look what you have done.......')
        print('to get current cam-position press key c')
        self.plotter.show()

    def add_text(self, **kwargs):
        r"""
        """
        self.plotter.add_text(**kwargs)
        self.text_objects.append(dict(**kwargs))

    def __call__(self, outpath: Path = Path.cwd() / 'spin.png') -> None:
        r"""
        Saves the image to a file

        Args:
            outpath(Path): output path for the png image created.
        """
        self._make_plotter(offscreen=True)
        for text_object in self.text_objects:
            self.plotter.add_text(**text_object)
        self.plotter.window_size = [4000, 4000]
        self.plotter.screenshot(str(outpath))
